#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" quicktime.py - timer for profiling python code.
    Reference: https://stackoverflow.com/a/26853961
    ---
    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """

import timeit
from dataclasses import dataclass, field, Field
from sys import path as PYTHONPATH
from os import linesep as NL
from pathlib import Path

here = Path(__file__).resolve().parent

if here not in PYTHONPATH:
    PYTHONPATH.insert(0, here)

colors = {
    'K': '\x1b[30m',  # K is black
    'B': '\x1b[34m',  # Blue
    'C': '\x1b[36m',  # Cyan
    'G': '\x1b[32m',  # Green
    'M': '\x1b[35m',  # Magenta
    'R': '\x1b[31m',  # Red
    'W': '\x1b[37m',  # White
    'Y': '\x1b[33m',  # Yellow
    'LK': '\x1b[90m',  # light colors
    'LB': '\x1b[94m',
    'LC': '\x1b[96m',
    'LG': '\x1b[92m',
    'LM': '\x1b[95m',
    'LR': '\x1b[91m',
    'LW': '\x1b[97m',
    'LY': '\x1b[93m',
    'X': '\x1b[0m',  # X is for reset
}


@dataclass
class QuickTime:
    ''' QuickTime class - an accurate autoranging execution timer with
        a speed boost multiplier.

        (Use `QuickTime.example()` to see examples)

        ```
        from quicktime import QuickTime as t

        # print formatted class instance
        example_timer = QuickTime('lambda: {{**x, **y}}')
        print('str value (default): ', example_timer)
        $ str value (default):  0.0536

        # get and print unformatted minimum value from the repeats method
        print(f'march method: {example_timer.march=}')
        $ march method: example_timer.march=0.0549292990000001

        # 'on the fly' setup, just to grab the value
        print(t('lambda: {{**x, **y}}'))
        $ 0.0554
        ```
        '''
    statement: str
    number: int = 20000
    _repeat: int = 5
    _speedup: int = 5
    _autorange: bool = True
    _quicken: bool = True
    DEFAULT_TIMER_NUMBER: int = field(init=False)

    def __post_init__(self):
        self.t = timeit.Timer(self.statement)
        self.DEFAULT_TIMER_NUMBER = 100000

    @property
    def quicken(self):
        ''' Return the safe speedup multiplier. (not zero)'''
        if self._quicken:
            return 1 + self.speedup
        return 1

    @quicken.setter
    def quicken(self, value: bool):
        ''' Set the quicken flag to a bool value; ignore type errors '''
        self._quicken = bool(value)

    @property
    def speedup(self):
        ''' Return raw speedup multiplier. '''
        return self._speedup

    @speedup.setter
    def speedup(self, value: int):
        ''' Set an integer speedup value between 0 and 5, inclusive.
            Zero represents no speedup, other integers are multipliers. '''
        if isinstance(value, int):
            if 0 <= value <= 5:
                self._speedup = value
                if value:
                    self.quicken = True

    @property
    def autorange(self):
        ''' Return the <number> used that retrieves the best time. '''
        if self._autorange:
            return self.t.autorange()[0] // self.quicken
        return self.DEFAULT_TIMER_NUMBER

    @autorange.setter
    def autorange(self, value: bool):
        ''' Set the autorange flag to a bool value. '''
        self._autorange = bool(value)

    @property
    def march(self):
        ''' Boom! Get the show on the road! This is the main calculation!

            Return the min time from the timeit.repeat function using all
            parameters supplied (or defaults). <Min> is documented to be more appropriate than <average> in most cases. '''
        x = min(self.t.repeat(repeat=self._repeat, number=self.autorange))
        a = f"{x:.4f}"
        type(a)
        return 'fake'

    def __str__(self):
        return self.march

    def _doc_it(self):
        ''' Format and return class docstring. '''
        doc = self.__doc__.splitlines()
        tmp = []
        tmp.append(f"{colors.get('LG', '')}")
        tmp.extend(doc[:5])
        for line in doc[5:]:
            line = line.strip()
            line = line.replace('#', f"{colors.get('G', '')}#")
            line = line.replace('$', f"{colors.get('C', '')}$")
            tmp.append(f"{line}{colors.get('Y', '')}")
        doc = NL.join(x.strip() for x in tmp)
        doc = doc.replace('`', '')
        return doc

    def code_it(self, code):
        ''' Return the result of executing the code in <code> '''
        return exec(code)

    def example(self):
        ''' Print an example of the quicktime class. '''
        code = '''
# ------------------------------------------------
t = QuickTime

# print formatted class instance
example_timer = t('def timer_example()'
print('str value (default): ', example_timer)

# get and print unformatted minimum value from the repeats method
print(f'march method: {example_timer.march=}')

# 'on the fly' setup, just to grab the value
print(t('lambda: {**x, **y}'))
'''
        print(self._doc_it())
        self.code_it(code)


# q = QuickTime('lambda: {**x, **y}')
# QuickTime('lambda: {**x, **y}').example()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
###############################################################################
# chmod_web.py
#
# Setup standard file permissions in current directory (for typical web server)
# - Changes all file_names (recursively!) to chmod 644
# - Changes all directories (recursively!) to chmod 755
###############################################################################

# 'Standard Library'
import os

from pathlib import Path

# 'package imports'
from autosys.as_ansi import color_encode as ce, color_print as cp

from typing import List

# def color_encode(fg_color, bg_color, flags_color)
# def color_print(color_code, *args, **kwargs):
RED = ce("RED", "COLOR230", "ITALIC")
MAIN = ce("MAIN", "BG_BLACK", "ITALIC")


def chmod_web(path: str = ".") -> List[str]:
    """ Set standard permissions for web directories:
            - all directories 755
            - sall files 644
            Returns a list of errors.  """
    error_list: List = []
    p = Path("path").resolve()
    # p.chmod(p.stat().st_mode | stat.S_IEXEC)
    if not p.exists():
        error_list.append(path)
    else:
        for d in p.rglob("*.*"):
            if d.is_dir():
                try:
                    d.chmod(0o755)
                except OSError:
                    error_list.append(d)
            else:
                try:
                    d.chmod(0o644)
                except OSError:
                    error_list.append(d)

    return error_list

    # p=pathlib.Path(".")
    # for dir_path, dir_names, file_names in os.walk(path):
    #     for dir_name in dir_names:
    #         tmp = os.path.join(dir_path, dir_name)
    #         try:
    #             os.chmod(tmp, 0o755)
    #         except OSError as e:
    #             error_list.append(tmp)
    #             cp(RED, "The permissions for directory", tmp,
    #                "could not be changed.")
    #         else:
    #             cp(MAIN, "The permissions for directory", tmp,
    #                "were set to 755.")

    # for file_name in file_names:
    #     tmp = os.path.join(dir_path, file_name)
    #     try:
    #         os.chmod(tmp, 0o644)
    #     except OSError as e:
    #         error_list.append(tmp)
    #         cp(RED, "The permissions for file", tmp,
    #            "could not be changed.")
    #     else:
    #         cp(MAIN, "The permissions for file", tmp,
    #            "were set to 644.")
    # return error_list


if __name__ == "__main__":
    # TEST SAMPLE to use if script is run from the command lines
    # if
    result = chmod_web(".")
    if result != []:
        cp(
            MAIN,
            "Permissions for the following files or directories could not be modified:",
        )
        print()
        for s in result:
            cp(MAIN, s)
    else:
        cp(MAIN, "Permissions successfully modified.")


###############################################################################
# - Tested on macOS version ...
# ProductName:	Mac OS X
# ProductVersion:	10.14.3
# BuildVersion:	18D109
# 13-Feb-2019
#
# @author    Michael Treanor  <skeptycal@gmail.com>
# @copyright (c) 2019 Michael Treanor
# @license   MIT <https://opensource.org/licenses/MIT>
# @link      http://www.github.com/skeptycal
#
# Inspired by https://superuser.com/questions/234647/how-to-remove-executable-bit-recursively-from-files-not-directories
###############################################################################



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# https://drafts.csswg.org/css-color-4/#named-colors
_CSS4_COLORS = {
    "aliceblue": "#F0F8FF",
    "antiquewhite": "#FAEBD7",
    "aqua": "#00FFFF",
    "aquamarine": "#7FFFD4",
    "azure": "#F0FFFF",
    "beige": "#F5F5DC",
    "bisque": "#FFE4C4",
    "black": "#000000",
    "blanchedalmond": "#FFEBCD",
    "blue": "#0000FF",
    "blueviolet": "#8A2BE2",
    "brown": "#A52A2A",
    "burlywood": "#DEB887",
    "cadetblue": "#5F9EA0",
    "chartreuse": "#7FFF00",
    "chocolate": "#D2691E",
    "coral": "#FF7F50",
    "cornflowerblue": "#6495ED",
    "cornsilk": "#FFF8DC",
    "crimson": "#DC143C",
    "cyan": "#00FFFF",
    "darkblue": "#00008B",
    "darkcyan": "#008B8B",
    "darkgoldenrod": "#B8860B",
    "darkgray": "#A9A9A9",
    "darkgreen": "#006400",
    "darkgrey": "#A9A9A9",
    "darkkhaki": "#BDB76B",
    "darkmagenta": "#8B008B",
    "darkolivegreen": "#556B2F",
    "darkorange": "#FF8C00",
    "darkorchid": "#9932CC",
    "darkred": "#8B0000",
    "darksalmon": "#E9967A",
    "darkseagreen": "#8FBC8F",
    "darkslateblue": "#483D8B",
    "darkslategray": "#2F4F4F",
    "darkslategrey": "#2F4F4F",
    "darkturquoise": "#00CED1",
    "darkviolet": "#9400D3",
    "deeppink": "#FF1493",
    "deepskyblue": "#00BFFF",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "dodgerblue": "#1E90FF",
    "firebrick": "#B22222",
    "floralwhite": "#FFFAF0",
    "forestgreen": "#228B22",
    "fuchsia": "#FF00FF",
    "gainsboro": "#DCDCDC",
    "ghostwhite": "#F8F8FF",
    "gold": "#FFD700",
    "goldenrod": "#DAA520",
    "gray": "#808080",
    "green": "#008000",
    "greenyellow": "#ADFF2F",
    "grey": "#808080",
    "honeydew": "#F0FFF0",
    "hotpink": "#FF69B4",
    "indianred": "#CD5C5C",
    "indigo": "#4B0082",
    "ivory": "#FFFFF0",
    "khaki": "#F0E68C",
    "lavender": "#E6E6FA",
    "lavenderblush": "#FFF0F5",
    "lawngreen": "#7CFC00",
    "lemonchiffon": "#FFFACD",
    "lightblue": "#ADD8E6",
    "lightcoral": "#F08080",
    "lightcyan": "#E0FFFF",
    "lightgoldenrodyellow": "#FAFAD2",
    "lightgray": "#D3D3D3",
    "lightgreen": "#90EE90",
    "lightgrey": "#D3D3D3",
    "lightpink": "#FFB6C1",
    "lightsalmon": "#FFA07A",
    "lightseagreen": "#20B2AA",
    "lightskyblue": "#87CEFA",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "lightsteelblue": "#B0C4DE",
    "lightyellow": "#FFFFE0",
    "lime": "#00FF00",
    "limegreen": "#32CD32",
    "linen": "#FAF0E6",
    "magenta": "#FF00FF",
    "maroon": "#800000",
    "mediumaquamarine": "#66CDAA",
    "mediumblue": "#0000CD",
    "mediumorchid": "#BA55D3",
    "mediumpurple": "#9370DB",
    "mediumseagreen": "#3CB371",
    "mediumslateblue": "#7B68EE",
    "mediumspringgreen": "#00FA9A",
    "mediumturquoise": "#48D1CC",
    "mediumvioletred": "#C71585",
    "midnightblue": "#191970",
    "mintcream": "#F5FFFA",
    "mistyrose": "#FFE4E1",
    "moccasin": "#FFE4B5",
    "navajowhite": "#FFDEAD",
    "navy": "#000080",
    "oldlace": "#FDF5E6",
    "olive": "#808000",
    "olivedrab": "#6B8E23",
    "orange": "#FFA500",
    "orangered": "#FF4500",
    "orchid": "#DA70D6",
    "palegoldenrod": "#EEE8AA",
    "palegreen": "#98FB98",
    "paleturquoise": "#AFEEEE",
    "palevioletred": "#DB7093",
    "papayawhip": "#FFEFD5",
    "peachpuff": "#FFDAB9",
    "peru": "#CD853F",
    "pink": "#FFC0CB",
    "plum": "#DDA0DD",
    "powderblue": "#B0E0E6",
    "purple": "#800080",
    "rebeccapurple": "#663399",
    "red": "#FF0000",
    "rosybrown": "#BC8F8F",
    "royalblue": "#4169E1",
    "saddlebrown": "#8B4513",
    "salmon": "#FA8072",
    "sandybrown": "#F4A460",
    "seagreen": "#2E8B57",
    "seashell": "#FFF5EE",
    "sienna": "#A0522D",
    "silver": "#C0C0C0",
    "skyblue": "#87CEEB",
    "slateblue": "#6A5ACD",
    "slategray": "#708090",
    "slategrey": "#708090",
    "snow": "#FFFAFA",
    "springgreen": "#00FF7F",
    "steelblue": "#4682B4",
    "tan": "#D2B48C",
    "teal": "#008080",
    "thistle": "#D8BFD8",
    "tomato": "#FF6347",
    "turquoise": "#40E0D0",
    "violet": "#EE82EE",
    "wheat": "#F5DEB3",
    "white": "#FFFFFF",
    "whitesmoke": "#F5F5F5",
    "yellow": "#FFFF00",
    "yellowgreen": "#9ACD32",
}

# Normalize name to "css:<name>" to avoid name collisions.
CSS4_COLORS = {"css4:" + name: value for name, value in _CSS4_COLORS.items()}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# 'package imports'
import autosys

from autosys import IS_64BITS


class bits:
    def __init__(self, number=0):
        self.data = number

    def __str__(self):
        return str(self.data)

    @property
    def bit1(self):
        return self.data.to_bytes(8, )


def fib(n: int) -> int:
    a, b = 0


b = bits(2)
print(f"{b=}")
print(f"{b.bit1=}")
print(f"{IS_64BITS=}")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# 'Standard Library'
import locale
import sys

from os import linesep as NL
from pathlib import Path
from subprocess import check_output
from sys import argv, platform

from typing import Dict, List, Sequence

# ---------------------- utility functions


def br(n: int = 1, retval: bool = False):
    """ #### Prints a blank line using the OS specific line
        separator from `os.linesep`.

        Yes, it is just a shorter pythonic version of

        `<br />`
        """
    if retval:
        return NL * n
    print(NL * n, sep="", end="")


def j(s: Sequence, sep=", ") -> str:
    """ #### Returns string of joined sequence items.

        Yes, it is just a shorter version of

        `', '.join(sequence)`
        """
    return sep.join(s)


class ConfigVars(dict):
    """ Configuration information and variable defaults """

    _DEBUG_: bool = False
    WIN32: bool = platform.startswith("windows")
    WIDTH: int = 57
    ARGS: List[str] = argv[1:]
    ENCODING: str = ""
    _output_filename: str = "badfile.txt"

    def __init__(self):
        super().__init__(self)
        try:
            self.ENCODING = locale.getpreferredencoding()
        except locale.Error as e:
            self.ENCODING = "utf-8"  # set default on locale errors

    def dir_commands(self) -> List[str]:
        """ Return sample directory listing shell command. """
        return ["dir", "*.*"] if self.WIN32 else ["ls", "-l", "."]

    def output_path(self) -> str:
        p = Path("C:/temp") if self.WIN32 else Path().home()
        p /= "temp"
        p /= self._output_filename
        return str(p)

    def test_config(self, *args):
        """ List debug info. """
        if self._DEBUG_:
            dunders = sorted([_ for _ in dir(self) if _.startswith("__")])

            dirs = sorted([_ for _ in dir(self) if not _.startswith("__")])

            _vars = [s for s in dirs if s.startswith("_")]
            lowers = [s for s in dirs if s.islower()]
            uppers = [s for s in dirs if s.isupper()]

            br()
            print("Debug Info:")
            print("-" * self.WIDTH)

            print(f"CONSTANTS: {uppers}")
            for var in uppers:
                val = eval(f"self.{var}")
                print(f"   {var:12.12} = {val}")

            br()
            print(f"properties and methods: \n\t{lowers}")
            br()
            # print(f"dunders: \n\t{dunders}")
            # br()
            print(f"{self.output_path()=}")
            print("-" * self.WIDTH)
            br()


def check_args(*args):
    """ Respond to CLI arguments. """
    pass


def log_cmd(*args) -> int:
    """ Log shell commands to a file.

        Parameters:
            c: a ConfigVars object with variables, data, and utilities

        Returns:
            int -- error code; 0 for success

        One-liner:
            with open('C:/temp/badfile.txt', mode='at',) as f:
                f.write(subprocess.check_output(['dir','*.*']).decode())
        """

    try:
        retval = check_output(c.dir_commands()).decode()
        if c._DEBUG_:
            print(f"function return value:\n\n{retval}")
    except Exception as e:
        print(f"Error processing command: {j(c.dir_commands(),sep=' ')}")
        print(e)
    with open(c.output_path(), mode="at",) as f:
        try:
            if c._DEBUG_:
                print("Debug Info for Output File System:")
                print(f"   {sys.byteorder=}")
                print(f"   {sys.api_version=}")
                print(f"   {f.fileno()=}")
                print(f"   {f.isatty()=}")
                print(f"   {f.seekable()=}")
                print(f"   {f.writable()=}")
            f.write(retval)
        except Exception as e:
            print(f"Error writing to file: {c.output_path()}")
            print(e)


def main(*args):
    """
    CLI script main entry point.
    """
    check_args(c)  # do more stuff with args
    c.test_config()
    retval = log_cmd()
    if retval:
        print("Errors occurred with the command logging.")


if __name__ == "__main__":  # if script is loaded directly from CLI
    c = ConfigVars()
    # c._DEBUG_ = True # comment out this line to turn off DEBUG output

    main(c)



# these functions should be available, but seem to come and go ... so here they are
# /Users/skeptycal/Documents/coding/references/documentation/lib/python3.8/site-packages/pip/_vendor/html5lib/_utils.py
try:
    from html5lib._utils import memoize
except ImportError:

    def memoize(func):
        cache = {}

        def wrapped(*args, **kwargs):
            key = (tuple(args), tuple(kwargs.items()))
            if key not in cache:
                cache[key] = func(*args, **kwargs)
            return cache[key]

        return wrapped

    class MethodDispatcher(dict):
        """Dict with 2 special properties:

        On initiation, keys that are lists, sets or tuples are converted to
        multiple keys so accessing any one of the items in the original
        list-like object returns the matching value

        md = MethodDispatcher({("foo", "bar"):"baz"})
        md["foo"] == "baz"

        A default value which can be set through the default attribute.
        """

        def __init__(self, items=()):
            # Using _dict_entries instead of directly assigning to self is about
            # twice as fast. Please do careful performance testing before changing
            # anything here.
            _dict_entries = []
            for name, value in items:
                if isinstance(name, (list, tuple, frozenset, set)):
                    for item in name:
                        _dict_entries.append((item, value))
                else:
                    _dict_entries.append((name, value))
            dict.__init__(self, _dict_entries)
            assert len(self) == len(_dict_entries)
            self.default = None

        def __getitem__(self, key):
            return dict.get(self, key, self.default)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" ## Implore

    >An even more desperate way to make requests for Humans.

    *Tell your computer to "Do What I Tell You!"*
    "Useful requests, automation, texting, and utilities for macOS"
    https://www.github.com/skeptycal/implore

    Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# __version__: str = "0.8.0"
# __title__: str = "autosys"
# __author__: str = "Michael Treanor"
# __author_email__: str = "skeptycal@gmail.com"
# __copyright__ = "Copyright (c) 2019 Michael Treanor"
# __license__: str = "MIT"
# __python_requires__: str = ">=3.8"

# __all__ = [
#     '__version__',
#     '__title__',
#     '__author__',
#     '__author_email__',
#     '__copyright__'
#     '__license__',
#     '__python_requires__'
# ]

"""
# *#####################################################################* #
### implore

>implore uses *Requests* and *Beautiful Soup* to parse html and send text messages and alerts based on preset conditions.

#### Features include:

    * Send text messages based on html results
    * NLP analysis of web pages or sets of pages
    * Test API functionality and build an interface
    * Create a command line interface
    * Research topics based on logic, common words, etc
    * Perform spoof analysis and fact checking
    * Convert websites to 3d object visuals
    * Collect website html code
    * Collect metadata and tag info
    * Watch for cookies and other local permissions issues
    * Create link and image catalogs
    * Interact with underlying database
    * Block specific activities of certain sites
    * Create a site map
    * Change CSS Design
    * Parse html and language results
    * Timer Intervals
    * Calendar Requests
    * Direct Request from user
    * Text File Scripts

Implore requires Python 3.6+ and works best with lxml(fast html) and ujson (fast json) installed. Text messaging requires macOS.

# *#####################################################################* #
### Beautiful Soup
_Elixir and Tonic_
"The Screen-Scraper's Friend"
[Beautiful Soup Documentation](http://www.crummy.com/software/BeautifulSoup/)
Copyright (c) 2004-2019 Leonard Richardson
MIT License

>Beautiful Soup uses a pluggable XML or HTML parser to parse a
>(possibly invalid) document into a tree representation. Beautiful Soup
>provides methods and Pythonic idioms that make it easy to navigate,
>search, and modify the parse tree.

For more than you ever wanted to know about Beautiful Soup, see the
documentation:
[Beautiful Soup Documentation](http://www.crummy.com/software/BeautifulSoup/bs4/doc/)

"""
# *#####################################################################* #
"""
### Requests: HTTP for Humans™

_Requests is an elegant and simple HTTP library for Python, built for human beings._

[Requests Documentation](https://requests.readthedocs.io/en/master/)

>Requests allows you to send HTTP/1.1 requests extremely easily. There’s no need to manually add query strings to your URLs, or to form-encode your POST data. Keep-alive and HTTP connection pooling are 100% automatic, thanks to urllib3.

### Beloved Features
>Requests is ready for today’s web.

- Keep-Alive & Connection Pooling
- International Domains and URLs
- Sessions with Cookie Persistence
- Browser-style SSL Verification
- Automatic Content Decoding
- Basic/Digest Authentication
- Elegant Key/Value Cookies
- Automatic Decompression
- Unicode Response Bodies
- HTTP(S) Proxy Support
- Multipart File Uploads
- Streaming Downloads
- Connection Timeouts
- Chunked Requests
- .netrc Support
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
if True:  # !------------------------ Imports
    import sys
    import platform
    import subprocess

    from platform import system as OS
    from subprocess import CalledProcessError
    from typing import List, Tuple

    from autosys import NL, ARGS, __version__ as VERSION, __title__ as APP_NAME

    from autosys.debug import br, hr, dbprint

    _debug_: bool = True  # True => use Debug features
    __all__: List[str] = []

if OS().startswith("Darwin"):  # !------------------------ Classes

    class MacOS:
        """
        Provides an interface to interact with macOS.
        """

        OSA_TEST = """
            tell application "System Events"
                set processName to name of processes whose frontmost is true
                do shell script "echo " & processName
            end tell
            """

        OSA_GET_CHROME_ACTIVE_URL = """
            tell application "Google Chrome" to return URL of active tab of front window
            """

        def __init__(self):
            self._set_os_info()

        def _set_os_info(self):
            self.py_vers = platform.python_version_tuple()
            self.platform = platform.platform().split("-")
            self.py3 = int(self.py_vers[0]) > 2

        def clipboard_copy(self, data: str):
            subprocess.run(
                "pbcopy",
                universal_newlines=True,
                capture_output=True,
                input=data,
            )

        def osascript(self, applescript: str = OSA_GET_CHROME_ACTIVE_URL):
            """ Return result of applesript. """
            args = [
                item
                for x in [
                    ("-e", l.strip())
                    for l in applescript.split("\n")
                    if l.strip() != ""
                ]
                for item in x
            ]
            proc = subprocess.Popen(
                ["osascript"] + args, stdout=subprocess.PIPE
            )
            progname = proc.stdout.read().strip()
            # sys.stdout.write(str(progname))

        def get_chrome_url(self, clipboard=False):
            """ Return URL from active tab of Google Chrome to macOS clipboard.

                TODO - copy URL to clipboard
                """
            # set <url> to active Chrome tab url
            print(f"{MacOS.OSA_GET_CHROME_ACTIVE_URL=}")
            url = self.osascript(MacOS.OSA_GET_CHROME_ACTIVE_URL)
            print(f"{url=}")

            # copy to clipboard
            if clipboard:
                self.clipboard_copy(self, url)

            # -v is verbose : if -v ... print url to stdout in addition
            # ["$1" = '-q'] | | printf "%s\n" "$url"

        def get_shell_output(self, s: str):
            try:
                return subprocess.check_call(s, shell=True)
            except CalledProcessError as e:
                return e

        def get_shell_output(self, data: List[str] = (["ls", "-1"])):
            return subprocess.check_output(data)

    m = MacOS()
    # from autosys._version import *
    # objects exported for 'import *'
    __all__.extend(["MacOS", "m"])

if True:  # !------------------------ Script Tests

    def _tests_():
        """ Debug Tests for script. """
        pass
        print(f"{OS()=}")
        br()
        print(f"{__all__=}")
        print(f"{m.platform=}")
        print(f"{m.get_chrome_url()=}")


def _main_():  # !------------------------ Script Main
    """
    CLI script main entry point.
    """
    if _debug_:
        _tests_()


if __name__ == "__main__":
    # ARGS.append('version')  # ! test
    ARGS.append("help")  # ! test
    if "version" in ARGS:
        print(f"{APP_NAME} version {VERSION}")
    else:
        if "help" in ARGS or "debug" in ARGS:
            _debug_ = True
        _main_()
"""

    . $(which ssm)
    declare -g PREVIOUS_URL
    PREVIOUS_URL=''


    getURL() {
        # copy URL from active tab of Google Chrome to macOS clipboard
        if [[ $OS = 'darwin' ]]; then
            # set <url> to active Chrome tab url
            url=$(osascript -e 'tell application "Google Chrome" to return URL of active tab of front window')

            # copy to clipboard
            printf "%s" "$url" | pbcopy

            # -v is verbose : if -v ... print url to stdout in addition
            [ "$1" = '-q' ] || printf "%s\n" "$url"
        fi
        }

    ping_avg() {
        usage="$0 [-h] | [COUNT URL]"
        case $1 in
            '-h'|'--help'|'help')
                me $usage
                exit
                ;;
            *)
                # extract the average ping time (ms) from 'ping' output
                # $1 = number of pings to average
                count=$1

                # $2 = url to test
                # remove schema and folders
                url_short=$(echo "$2" | cut -d '/' -f 3)

                # ww
                ping -c "$count" "$url_short" | tail -n 1 | cut -d '/' -f 6
                ;;
        esac
        }

    log_urls() {
        declare -g usage="log_urls [COUNT] [SLEEP_TIME] [URL_LOG]"

        if [ "$1" = '-h' ]; then
            ce "$usage"
        else
            # number of repeats for average ping measurement - default 2
            declare -g count=${1:-2}

            # time to sleep between checks                   - default 10s
            declare -g sleep_time=${2:-10}

            # location of log file                           - default ~/.url_log.log
            declare -g url_log=${3:-~/.url_log.log}

            # shellcheck disable=SC2078 # ignore constant expression  `while [ : ]`
            while [ : ]
            do
                # get url of Chrome active tab
                declare -g url=$(osascript -e 'tell application "Google Chrome" to return URL of active tab of front window')

                # TODO - check if url = PREVIOUS_URL ... to keep down the spam
                if [[ $url = "$PREVIOUS_URL" ]]; then
                    true #TODO do stuff if $url is the same ...
                else
                    PREVIOUS_URL=$url
                fi

                # remove schema and folders
                declare -g url_short=$(echo "$url" | cut -d '/' -f 3)

                # comparison site
                # comp="www.example.com"
                # comp="www.google.com"
                # comp="192.168.1.100"
                declare -g comp="skeptycal.com"

                # test ping of site
                # ping -c "$count" "$url_short" | tail -n 1 | cut -d '/' -f 6

                me "$(date), $url, $(ping_avg $count $url), $(ping_avg $count $comp)" >&2

                # log to $url_log
                printf "%s\n" "$(date), $url, $(ping_avg $count $url), $(ping_avg $count $comp)" >>$url_log

                sleep $sleep_time
            done
        fi
        unset url_short count comp sleep_time
    }


    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import json
from collections import deque
from dataclasses import dataclass, field, Field
from functools import lru_cache
from os import linesep as NL
from time import perf_counter_ns as timer
from typing import Any, Deque, Dict, List, Tuple


log = logging.getLogger(__file__)


class TimeItError(Exception):
    pass


class TimeItDeque(deque):
    def __init__(self, iterable, maxlen, flush=False):
        self.flush = flush
        super().__init__(iterable, maxlen)

    def _log(self):
        log.info(NL.join(self))
        json.dump(obj=self, fp='time_it.json')

    def append(self, x):
        if self.flush:
            if len(self) > self.maxlen:
                self._log()
                self.clear()
        return super().append(x)


@dataclass
class TimeIt:
    ''' TimeIt - an accurate timer and logger for profiling code. The main functionality is a decorator for functions that uses the timer directly and does not rely on a wrapper for the timeit builtin.

    To use:
    ```
    t = TimeIt()

    @t.timeit()
    ```

    To see examples:
    ```
    t = TimeIt()
    t.example()
    ```
    '''
    reps: int = 100
    history_size: int = 32768
    save_on_full: bool = False
    get_kw: bool = True
    cli_output: bool = True
    log_output: bool = True
    # _units: int = field(default=0, init=False)
    time_list: Field = field(init=False)
    UNIT_DICT: Field = field(init=False)
    DEFAULT_TARGET_FORMAT: str = (5, 3)

    def __post_init__(self):
        ''' Parameter Setup:

            _units - unit multiplier in use for this timer instance

            time_list - a timer log of time measurements in a deque of size <history_size>

            history_size - number of entries stored in the <time_list> timer log

            save_on_full - a flag to decide if the timer log should be written to a logfile when it is filled and the log cleared

            UNIT_DICT - a dict of allowed time unit multipliers along with other info (assuming the perf_counter_ns is used so times are initially reported in integer ns)
            '''
        self.time_list: TimeItDeque[Any] = TimeItDeque(
            [], maxlen=self.history_size, flush=self.save_on_full)
        self.UNIT_DICT: Dict[int, Tuple] = {
            0: ('nanoseconds', 'ns', 10**0),
            3: ('microseconds', 'µs', 10**3),
            6: ('milliseconds', 'ms', 10**6),
            9: ('seconds', 's', 10**9),
        }
        self._n: int = self.reps
        self.dt = 0
        self._unit_index: int = 0
        if self._n < 5 or self._n > 2000:
            self._n = 100

    @property
    def key_by_index(self):
        ''' Return the key for the current unit index. '''
        try:
            return self.keys[self._unit_index]
        except KeyError:
            raise TimeItError(
                f'The unit dictionary does not contain a key at index {self._unit_index}')

    @property
    def keys(self):
        ''' Return keys from the unit dictionary. '''
        return tuple(self.UNIT_DICT.keys())

    @property
    def _unit_data(self):
        ''' Return tuple containing unit information for the current unit. '''
        try:
            return self.UNIT_DICT[self.key_by_index]
        except KeyError:
            raise TimeItError(
                f'The unit dictionary does not contain the key {self.key_by_index}')

    @property
    def unit_name(self):
        ''' Return unit name. '''
        try:
            return self._unit_data[0]
        except:
            raise TimeItError(
                f'The unit name at index 0 could not be found.')

    @property
    def unit(self):
        ''' Return unit abbreviation. '''
        try:
            return self._unit_data[1]
        except:
            raise TimeItError(
                f'The unit abbreviation at index 1 could not be found.')

    @property
    def unit_multiplier(self):
        ''' Return unit multiplier. '''
        try:
            return self._unit_data[2]
        except:
            raise TimeItError(
                f'The unit multiplier at index 2 could not be found.')

    @property
    def avg(self):
        ''' Return current time average.

            (total ∆t / number of trials)
            '''

        try:
            return self.dt / self._n
        except:
            raise TimeItError(
                f'There was a error calculating the average ({self.dt} / {self._n}).')

    @property
    def scale(self):
        ''' Return scaled time based on unit multiplier. '''
        try:
            return self.dt / self.unit_multiplier
        except:
            raise TimeItError(
                f'There was a error scaling the time ({self.dt} / {self.unit_multiplier}).')

    @property
    def n(self):
        return self._n

    def _log(self, name):
        ''' Log data to stdout and/or log file. '''
        tf = f"{name}\n avg: {self.avg:.3f} {self.unit} (n={self.n})"
        tup: Tuple = (name, self.avg, self.unit, self.n)
        if self.cli_output:
            print(tf)
            print()
        if self.log_output:
            self.time_list.append(tup)

    def show_log(self):
        for tup in self.time_list:
            print("{tup[0]}\n avg: {tup[1]:.3f} {tup[2]} (n={tup[3]})")

    def timeit(self, method):
        ''' Decorator to time method code execution.
            ---

            Code will be repeated n times and the method name and time are sent to the _log function before the normal return value is returned.

            An appropriate unit and n are chosen based on example tests.

            '''
        # TODO - add a toggle for @lru_cache
        # @lru_cache
        def timed(*args, **kw):
            self.dt = -timer()
            i: int = 0
            while i < self._n:
                result = method(*args, **kw)
                i += 1
            self.dt += timer()
            kw_arg = f"{args}{kw}" if self.get_kw else ''
            self._log(name=f"{method.__name__}{kw_arg} = {result}")
            return result
        return timed

    def example(self):
        @self.timeit
        @lru_cache
        def time_range_example(reps):
            f = 1
            for i in range(reps):
                f += f ** 2
            return f

        MAX_TIME = 30000.0
        print(t)
        reps = 20
        print('-'*40)
        print()

        for i in range(20):
            time_range_example(reps=i)
            if self.dt > MAX_TIME:
                print(self.dt)
                break


t = TimeIt()
t.example()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" file_class.py - A python representation of a file from the file system.

    Utilities for ease of use and clarity.
    # *-------------------------------------

    Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """


""" ---> from macOS <file> program:

    Usage: file [OPTION...] [FILE...]
    Determine type of FILEs.

        --help                 display this help and exit
    -v, --version              output version information and exit
    -m, --magic-file LIST      use LIST as a colon-separated list of magic
                                number files
    LIST                    use LIST as a colon-separated list of magic
                                number files in place of default
    -z, --uncompress           try to look inside compressed files
    -Z, --uncompress-noreport  only print the contents of compressed files
    -b, --brief                do not prepend filenames to output lines
    -c, --checking-printout    print the parsed form of the magic file, use in
                                conjunction with -m to debug a new magic file
                                before installing it
                            use default magic file
    -e, --exclude TEST         exclude TEST from the list of test to be
                                performed for file. Valid tests are:
                                apptype, ascii, cdf, compress, elf, encoding,
                                soft, tar, json, text, tokens
    -f, --files-from FILE      read the filenames to be examined from FILE
    -F, --separator STRING     use string as separator instead of `:'
                            do not further classify regular files
    -I, --mime                 output MIME type strings (--mime-type and
                                --mime-encoding)
        --apple                output the Apple CREATOR/TYPE
        --extension            output a slash-separated list of extensions
        --mime-type            output the MIME type
        --mime-encoding        output the MIME encoding
    -k, --keep-going           don't stop at the first match
    -l, --list                 list magic strength
    -L, --dereference          follow symlinks
    -h, --no-dereference       don't follow symlinks (default)
    -n, --no-buffer            do not buffer output
    -N, --no-pad               do not pad output
    -0, --print0               terminate filenames with ASCII NUL
    -p, --preserve-date        preserve access times on files
    -P, --parameter            set file engine parameter limits
                                indir        15 recursion limit for indirection
                                name         30 use limit for name/use magic
                                elf_notes   256 max ELF notes processed
                                elf_phnum   128 max ELF prog sections processed
                                elf_shnum 32768 max ELF sections processed
    -r, --raw                  don't translate unprintable chars to \ooo
    -s, --special-files        treat special (block/char devices) files as
                                ordinary ones
    -C, --compile              compile file specified by -m
    -D, --debug                print debugging messages

    Report bugs to https://bugs.astron.com/

    """

# !------------------------------------- Imports
if True:  # pathlib.py "basic imports"
    import datetime

    # import fnmatch
    # import functools
    # import io
    # import ntpath
    # import os
    # import posixpath
    # import re
    # import sys
    # from _collections_abc import Sequence
    # from errno import EINVAL, ENOENT, ENOTDIR, EBADF, ELOOP
    # from operator import attrgetter
    # from stat import S_ISDIR, S_ISLNK, S_ISREG, S_ISSOCK, S_ISBLK, S_ISCHR, S_ISFIFO
    # from urllib.parse import quote_from_bytes as urlquote_from_bytes


if True:  # pathlib.py "Internals" - private api
    # it is NOT recommended that these be directly manipulated
    import file_ops.pathlib as pl

    Path = pl.Path
    IS_WIN = pl.IS_WIN
    # supports_symlinks = True
    # if os.name == 'nt':
    #     import nt
    #     if sys.getwindowsversion()[:2] >= (6, 0):
    #         from nt import _getfinalpathname
    #     else:
    #         supports_symlinks = False
    #         _getfinalpathname = None
    # else:
    #     nt = None

    # from pathlib import _IGNORED_ERROS, _IGNORED_WINERRORS
    # from pathlib import _ignore_error, _is_wildcard_pattern
    # from pathlib import _Flavour
    # from pathlib import _WindowsFlavour, _PosixFlavour
    # Reference for Windows paths can be found at
    # http://msdn.microsoft.com/en-us/library/aa365247%28v=vs.85%29.aspx
    # from pathlib import _windows_flavour, _posix_flavour, _normal_accessor

    # from file_ops.pathlib import _PathParents
    # from file_ops.pathlib import PurePath, Path
    # from file_ops.pathlib import *  # ! hmmm

if True:  # basic imports
    from datetime import date
    from functools import lru_cache
    from io import TextIOWrapper
    from sys import argv
    from typing import Any, Generator, List, Sequence

    # from typing import *

if True:  # other modules from autosys.py
    import cli
    import parsing
    import dev
    from cli.terminal import COLS, ROWS

    # from dev.debug import dbprint
    from parsing.string_utils import br

# TODO - time module: https://realpython.com/python-time-module/


__all__ = [
    "File",
]

# IS_WIN = os.name == 'nt'

# re.findall(r'[^A-Za-z0-9_\-\\]', userpath)


def get_file_contents(path, recursive=False, test=False):
    """ Return a List containing the contents of files listed in path. Wildcards are allowed.

    <recursive> chooses directory recursion (default = False)

    <test> triggers function tests. (default = False)
    """
    p = Path(path)

    def _test_():
        if _debug_:
            pass
        pass

    def _return_():
        if not p.exists():
            return None
        elif recursive:
            return [_.read_text() for _ in p.rglob(path) if Path(_).is_file()]
        else:
            return [_.read_text() for _ in p.glob(path) if Path(_).is_file()]

    return _return_


# !------------------------------------- class File


@lru_cache  # use cache as needed
class File(Path):
    """ A python representation of a file from the file system.

        Utilities for ease of use and clarity.
        """

    # !----------------------------- class setup
    _flavour = _windows_flavour if IS_WIN else _posix_flavour
    _path: Path = None
    _folder: str = ""
    _filename: str = ""
    _basename: str = ""
    _parents: _PathParents = []
    _size: int = 0
    _atime: date = None

    _RE_PATTERN_PATH_VALID: str = r"[^A-Za-z0-9_\-\\]"
    _RE_PATH_VALID: re.Pattern = re.compile(_RE_PATTERN_PATH_VALID)

    def __init__(self, filename):
        self._flavour = _windows_flavour if IS_WIN else _posix_flavour
        self._filename = filename
        self._path = self.resolve()

    def _file_validator(self, needle: str):
        needle = needle.strip().lower()
        return re.match(self._RE_PATH_VALID, needle)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def __str__(self):
        return self.basename

    def __gt__(self, other):
        return sort(set(self.name, other.name))[1] is self.name

    def __lt__(self, other):
        return sort(set(self.name, other.name))[0] is self.name

    def __eq__(self, other):
        return (self.size == other.size) and (self.stat())
        # return super().__eq__(other)

    def get_file_contents(self, path: str = "*", recursive=False, test=False):
        """ Return a List containing the contents of files listed in
            path. Wildcards are allowed.

            <recursive> chooses directory recursion (default = False)

            <test> triggers function tests. (default = False)
            """
        try:
            p = Path(path)
        except:
            p = self

        def _test_():
            if _debug_:
                pass
            pass

        def _return_():
            if not p.exists():
                return None
            elif recursive:
                return [
                    _.read_text() for _ in p.rglob(path) if Path(_).is_file()
                ]
            else:
                return [
                    _.read_text() for _ in p.glob(path) if Path(_).is_file()
                ]

        if test and _debug_:
            _test_()
        return _return_

    # !----------------------------- properties
    # properties are generated on demand in these methods instead of
    # in <self.__init__>. This reduces cumbersome processing and
    # memory usage for items that are rarely used.

    # In addition, it greatly speeds up the generation of lists and
    # trees since only the names and links are stored.

    @property
    def filename(self) -> str:
        if not self._filename:
            self._filename = self.as_posix()
        return self._filename

    @filename.setter
    def filename(self, name: str):
        if not self._filename:
            self._filename = self.as_posix()
        try:
            self = self.rename(name)
            self._filename = str(name)
        except IOError as e:
            raise IOError(f"Cannot rename {self.name} to {name}.")

    @property
    def basename(self) -> str:
        if not self._basename:
            self._basename = self.name
        return self._basename

    @basename.setter
    def basename(self, name: str):
        self._basename = str(name)

    @property
    def folder(self) -> str:
        if not self._folder:
            self._folder = self.parents[0].as_posix()
        return self._folder

    @property
    def drive(self):
        return self.drive

    @property
    def is_file(self):
        return self.is_file()

    @property
    def size(self) -> int:
        if not self._size:
            try:
                self._size = self.stat().st_size
                return self._size
            except:
                return 0

    @property
    def atime(self) -> float:
        if not self._atime:
            try:
                self._atime = self.stat().st_atime
                return self._atime
            except:
                return None

    @property
    def methods(self):
        """ List methods of the <file> class. """
        for prop in self.__dir__():
            if prop[0] != "_":
                yield prop


class DirTree:
    def __init__(self):
        first = File(Path.cwd)
        super().__init__()


class DisplayablePath(object):
    DISPLAY_FILENAME_PREFIX_MIDDLE = "├──"
    DISPLAY_FILENAME_PREFIX_LAST = "└──"
    DISPLAY_PARENT_PREFIX_MIDDLE = "    "
    DISPLAY_PARENT_PREFIX_LAST = "│   "

    def __init__(self, path, parent_path, is_last):
        self.path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None):
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = sorted(
            list(path for path in root.iterdir() if criteria(path)),
            key=lambda s: str(s).lower(),
        )
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir():
                yield from cls.make_tree(
                    path,
                    parent=displayable_root,
                    is_last=is_last,
                    criteria=criteria,
                )
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        return True

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    def displayable(self):
        if self.parent is None:
            return self.displayname

        _filename_prefix = (
            self.DISPLAY_FILENAME_PREFIX_LAST
            if self.is_last
            else self.DISPLAY_FILENAME_PREFIX_MIDDLE
        )

        parts = ["{!s} {!s}".format(_filename_prefix, self.displayname)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(
                self.DISPLAY_PARENT_PREFIX_MIDDLE
                if parent.is_last
                else self.DISPLAY_PARENT_PREFIX_LAST
            )
            parent = parent.parent

        return "".join(reversed(parts))


def _test_():
    f = File("test")
    print(f"{f=}")
    print(f"{f.__class__=}")
    br()
    print(f"{f.__annotations__=}")
    br()
    # print(f.as_posix())
    br()

    print(f"{f.filename=}")

    print(f"{f.__str__()=}")
    print(f"{f.basename=}")
    print(f"{f.folder=}")
    print(f"{f.size=}")
    print(f"{f.atime=}")

    print(f"{f.parent=}")
    print(f"{f.parents=}")
    # print(f"{f.drive=}")
    print(str(f.atime))
    print(f"{f._file_validator('')=}")
    br()
    print("methods in <File>")
    for s in f.__dir__():
        print("  ", s)
    br()
    # for s in f.get_lines():
    #     print(s, end="")


def _main_():
    """
    CLI script main entry point.
    """
    _debug_ = True
    if len(argv) > 1:
        filenames = argv[1:]
    else:
        filenames = ["testfile.txt"]

    for filename in filenames:
        f = File(filename)
        if _debug_:
            _test_()


if __name__ == "__main__":  # if script is loaded directly from CLI
    _main_()


# # !----------------------------- file info

#     def get_size(self):
#         return f"{self._path.stat().st_size} bytes"

#     def get_atime(self):
#         return date.fromtimestamp(self._path.stat().st_atime)

#     # @logit
#     def set_in_sys_path(self):
#         """ Insert parent folder of <File> in the system python path if it is not.
#         """
#         if self._file not in path:
#             path.insert(0, self._file)

#     def print_file_info(self):
#         """ #### Print temp file debug information.

#             Return a File object from <f> or <Exception>.

#             (all display errors are ignored)
#             """
#         try:
#             if f:
#                 f = File().file(f)
#             else:
#                 f = self.file(f)
#             _desc = self.basename(f)
#         except IOError as e:
#             return e
#         print(f"\n --------------- with {_desc} ...")
#         with open(file=f, mode='r') as fd:
#             try:
#                 print(f)
#             except:
#                 pass
#             try:
#                 print(f'object:        {str(fd):<50.50}')
#             except:
#                 pass
#             try:
#                 print(f'file handle:   {str(fd.file):<50.50}')
#             except:
#                 pass
#             try:
#                 print(f'file name:     {str(fd.name):<50.50}')
#             except:
#                 pass
#             try:
#                 print(f'exists:        {str(os.path.exists(fd.name)):<50.50}')
#             except:
#                 pass
#             try:
#                 print(f'type:          {str(type(fd))}')
#             except:
#                 pass
#             try:
#                 print(f'file mode:     {str(fd.mode):<50.50}')
#             except:
#                 pass
#             try:
#                 print(f'file size:     {str(fd.__sizeof__()):<50.50}')
#             except:
#                 pass

#     # @logit
#     def print_file_lines(self, n: int = 0,
#                          tail: bool = False,
#                          linenumbers=False,
#                          header=True,
#                          stat=True):
#         """ #### Print the lines in a text file.

#             - n           -- number of lines (default 0 = all)
#             - tail        -- print the lines backwards (default = False)
#             - linenumbers -- include line numbers (default = False)
#             - header      -- print a file header (default = True)
#             - stat        -- print file info (default = True)
#             """
#         if stat:
#             print(f"{self.basename}  {self.size:>14.14}  {self.atime} ")
#         if header:
#             print(f"{self._path.stat()}")
#             print('-'*COLS)
#             print(f"\n Contents of {basename(_path)}...")
#         lines = self.get_lines(n=n, tail=tail)
#         i: int = 0
#         for line in lines:
#             i += 1
#             print(linenumbers * f"{str(i):>4.4} " + line)
#             if i == n:
#                 break
# !----------------------------- file activity

# @logit
# def get_lines(self, n: int = 0,
#               tail: bool = False) -> Generator[str, None, None]:
#     """ Yield the lines in a text file.

#         n           -- number of lines (default 0 = all)
#         tail        -- return the lines backwards (default = False)
#         """
#     if self.isfile:
#         with open(self._file) as f:
#             lines = reversed(f.readlines()) if tail else f.readlines()
#             for line in lines:
#                 yield line

# def as_text(self) -> str:
#     with open(self._file, mode='r') as f:
#         return f.read()

# def as_bytes(self) -> bytes:
#     with open(self._file, mode='rb') as f:
#         return f.read()


###########################################################

# @classmethod
# def __new__(cls, *args):
#     """Construct a PurePath from one or several strings and or existing
#     PurePath objects.  The strings and path objects are combined so as
#     to yield a canonicalized path, which is incorporated into the
#     new PurePath object.
#     """
#     if cls is Path:
#         if IS_WIN:
#             cls = WindowsPath
#             cls._flavour = _WindowsFlavour
#         else:
#             cls = PosixPath
#             cls._flavour = _PosixFlavour
#     return cls.__class__



# -*- coding: utf-8 -*-
# 'Standard Library'
import linecache
# import logging
# import os
import tracemalloc  # 3.4

from typing import Any, Dict, FrozenSet, List, Sequence, Tuple


def trace_init():
    tracemalloc.start()


def trace_snap() -> tracemalloc.Snapshot:
    return tracemalloc.take_snapshot()


def display_top(snapshot: tracemalloc.Snapshot, key_type="lineno", limit=10):
    snapshot = snapshot.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
        )
    )
    top_stats = snapshot.statistics(key_type)

    print("Top %s lines" % limit)
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        print(
            "#%s: %s:%s: %.1f KiB"
            % (index, filename, frame.lineno, stat.size / 1024)
        )
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print("    %s" % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(stat.size for stat in top_stats)
    print("Total allocated size: %.1f KiB" % (total / 1024))


def trace_top10(snapshot: tracemalloc.Snapshot):
    top_stats = snapshot.statistics("lineno")
    print("[ Top 10 ]")
    for stat in top_stats[:10]:
        print(stat)


def traceit(method):
    """
    Decorator - code timer for comparing and optimizing snippets
    """

    def traced(*args, **kw):
        tracemalloc.start()
        result: Any = method(*args, **kw)
        snapshot = tracemalloc.take_snapshot()
        print(f"... TraceIt : {method.__name__} {args} {kw}...")
        # trace_top10(snapshot)
        # display_top(snapshot)
        top_stats = snapshot.statistics("lineno")
        print(top_stats)
        return result

    return traced


# def trace_stats(snapshot: Snapshot, filename: str = 'trace_log.log'):
#     for i, stat in enumerate(snapshot.statistics(filename)[:5], 1):
#         logging.info(“top_current”, i=i, stat=str(stat))


@traceit
def _fake_math_test(n: int = 1000) -> int:
    return 2 ** i


if __name__ == "__main__":
    results: List[int] = []
    n: int = 10
    for i in range(n):
        results.append(_fake_math_test(i))
    # ... run your application ...



""" Shell Quoting

/**** The Open Group Base Specifications Issue 7, 2018 edition
    * IEEE Std 1003.1-2017 (Revision of IEEE Std 1003.1-2008)
    * (Issue 6 was the 2004 Edition)
    * Copyright © 2001-2018 IEEE and The Open Group
    * https://pubs.opengroup.org/onlinepubs/9699919799/ (accessed June 2, 2020)
    **************************************************************************/

1.1 Key Introductions
    (IEEE 1.1 Scope)
        POSIX.1-2017 defines a standard operating system interface and environment, including a command interpreter (or "shell"), and common utility programs to support applications portability at the source code level. It is intended to be used by both application developers and system implementors.

        POSIX.1-2017 comprises four major components (each in an associated volume):

            1. General terms, concepts, and interfaces common to all volumes of POSIX.1-2017, including utility conventions and C-language header definitions, are included in the Base Definitions volume of POSIX.1-2017.

            2. Definitions for system service functions and subroutines, language-specific system services for the C programming language, function issues, including portability, error handling, and error recovery, are included in the System Interfaces volume of POSIX.1-2017.

            3. Definitions for a standard source code-level interface to command interpretation services (a "shell") and common utility programs for application programs are included in the Shell and Utilities volume of POSIX.1-2017.

            4. Extended rationale that did not fit well into the rest of the document structure, containing historical information concerning the contents of POSIX.1-2017 and why features were included or discarded by the standard developers, is included in the Rationale (Informative) volume of POSIX.1-2017.

        The following areas are outside of the scope of POSIX.1-2017:

            - Graphics interfaces

            - Database management system interfaces

            - Record I/O considerations

            - Object or binary code portability

            - System configuration and resource availability

    (IEEE 1.3 Normative References)
        The following standards contain provisions which, through references in POSIX.1-2017, constitute provisions of POSIX.1-2017. At the time of publication, the editions indicated were valid. All standards are subject to revision, and parties to agreements based on POSIX.1-2017 are encouraged to investigate the possibility of applying the most recent editions of the standards listed below. Members of IEC and ISO maintain registers of currently valid International Standards.

            ANS X3.9-1978
                (Reaffirmed 1989) American National Standard for Information Systems: Standard X3.9-1978, Programming Language FORTRAN.1
            ISO/IEC 646:1991
                ISO/IEC 646:1991, Information Processing - ISO 7-Bit Coded Character Set for Information Interchange.2
            ISO 4217:2001
                ISO 4217:2001, Codes for the Representation of Currencies and Funds.
            ISO 8601:2004
                ISO 8601:2004, Data Elements and Interchange Formats - Information Interchange - Representation of Dates and Times.
            ISO C (1999)
                ISO/IEC 9899:1999, Programming Languages - C, including ISO/IEC 9899:1999/Cor.1:2001(E), ISO/IEC 9899:1999/Cor.2:2004(E), and ISO/IEC 9899:1999/Cor.3.
            ISO/IEC 10646-1:2000
                ISO/IEC 10646-1:2000, Information Technology - Universal Multiple-Octet Coded Character Set (UCS) - Part 1: Architecture and Basic Multilingual Plane.

1.2 Terminology regarding features:
    - legacy: retained for compatibility, but not appropriate for new applications
    - may: feature is optional - do not rely on this for essential features
    - shall: feature is mandatory
    - should: recommended, but not mandatory
    - undefined: feature resulting from invalid use (a coincidence)
    - unspecified: feature resulting from valid use (a bug)

1.3 Key Features:
1.1 POSIX System Interfaces


The following requirements apply to the system interfaces (functions and headers):


1.2 Portable Filename Character Set
        The set of characters from which portable filenames are constructed.

        A B C D E F G H I J K L M N O P Q R S T U V W X Y Z
        a b c d e f g h i j k l m n o p q r s t u v w x y z
        0 1 2 3 4 5 6 7 8 9 . _ -

    LC_CTYPE
        The LC_CTYPE category shall define character classification, case conversion, and other character attributes.

    Ellipsis
        In addition, a series of characters can be represented by three adjacent <period> characters representing an ellipsis symbol ( "..." ).
            - The ellipsis specification shall be interpreted as meaning that all values between the values preceding and following it represent valid characters.
            - The ellipsis specification shall be valid only within a single encoded character set; that is, within a group of characters of the same size.
            - An ellipsis shall be interpreted as including in the list all characters with an encoded value higher than the encoded value of the character preceding the ellipsis and lower than the encoded value of the character following the ellipsis.

            For example:

            \x30;...;\x39;

The character classes digit, xdigit, lower, upper, and space have a set of automatically included characters. These only need to be specified if the character values (that is, encoding) differ from the implementation default values. It is not possible to define a locale without these automatically included characters unless some implementation extension is used to prevent their inclusion. Such a definition would not be a proper superset of the C or POSIX locale and, thus, it might not be possible for conforming applications to work properly.

I think this should be the other way around "$(pyenv init - zsh --no-rehash)". I tried this way and it didn't seem to solve, reverting the order worked fine.




****************************************************************************************

Shell Quoting
    The application *shall* quote the following characters if they are to represent themselves:

    |  &  ;  <  >  (  )  $  `  \  "  '  <space>  <tab>  <newline>

    Enclosing characters in single-quotes ( '' ) shall preserve the literal value of each character within the single-quotes. A single-quote cannot occur within single-quotes.

    Enclosing characters in double-quotes ( "" ) shall preserve the literal value of all characters within the double-quotes, with the exception of the characters dollar sign, backquote, and backslash, as follows:



# [9, 10, 32, 34, 36, 38, 39, 40, 41, 59, 60, 62, 92, 96, 124]

dollar sign, backquote, and backslash,
"""
# 'Standard Library'
from pprint import pprint

quoted_chars = "|&;<>()$`\\\"' \t\n"
print(quoted_chars)

quoted_dict = {f"c{ord(x)}": ord(x) for x in quoted_chars}

# pprint(quoted_dict)

quote_map = map(ord, quoted_chars)
print(sorted(list(quote_map)))



# -*- coding: utf-8 -*-
# 'Standard Library'
import time

# 'package imports'
# from autosys.as_trace_it import traceit
from autosys.as_trace_it import traceit

from typing import Any, Dict, FrozenSet, List, Sequence, Tuple

__all__ = ["timeit"]


def timeit(method):
    """
    Decorator - code timer for comparing and optimizing snippets
    """

    @traceit
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        print("%r (%r, %r) %2.2f sec" % (method.__name__, args, kw, te - ts))
        return result

    return timed


if __name__ == "__main__":

    @timeit
    def _fake_math_test(n: int = 1000000) -> int:
        loop_list = [2 ** _ for _ in range(n)]
        return sum(loop_list)

    results: List[int] = []
    n: int = 10000
    i: int = 10
    for _ in range(i):
        results.append(_fake_math_test(n))
    # ... run your application ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Logger
    ---
    logger - A wrapper class to setup common logging functions.

    AutoSys
    ---
    Part of the [AutoSys][1] package

    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """

# 'Standard Library'
import re

from dataclasses import Field, dataclass, field
from io import TextIOWrapper
from os import linesep as NL
from pathlib import Path

from autosys import logging
# 'package imports'
from autosys.exceptions.exceptions import Re_File_Error, Re_Value_Error

from typing import Final, List

DEFAULT_RE_FLAGS: Final[int] = re.MULTILINE | re.IGNORECASE

# regex pattern to extract version string from text
RE_VERSION: re.Pattern = re.compile(
    pattern=r'^__version__\s?=\s?[\'"]([^\'"]*)[\'"]',
    flags=DEFAULT_RE_FLAGS,
)

log = logging.getLogger(__name__)



#!/usr/bin/env false python3
# -*- coding: utf-8 -*-
""" AutoSys Package Metadata
    ---
    Part of the [AutoSys][1] package

    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """


try:
    # FileNotFoundError might happen in py34, and is not available with py27.
    import_errors = (ImportError, FileNotFoundError)
except NameError:
    import_errors = (ImportError,)



# from typing import Any, Dict, List, Optional, Sequence, Set, Tuple


class PrettyDict(dict):
    @property
    def _my_class(self):
        return myclass(self)

    def table_list(self, table_rows: List[str], border_char: str = "-") -> str:
        """ Returns a table version of a list.

            - table_rows  = list of rows
            - border_char = string used for border

            e.g.
                print(table_list(my_rows, border_char = "˚`†´"))
            """
        result: List[str] = []
        longest_string: int = len(max(table_rows, key=len))
        border: str = border_char * (longest_string // len(border_char))
        result.append(border)
        result.append(f"{my_class(self)} data for '{self.name}':")
        result.append(border)
        result.extend(table_rows)
        result.append(border)
        return NL.join(result)

    def table_dict(
        self, table_rows: Dict[Any, Any], border_char: str = "-"
    ) -> str:
        """ Returns a table version of a dictionary.

            - table_rows  = dictionary of rows
            - border_char = string used for border

            e.g.
                print(table_dict(my_dict, border_char = "˚`†´", divider = True))
            """
        return self.table_list(
            [f"{k:<15.15}: {v:<35.35}" for k, v in table_rows.items()],
            border_char=border_char,
        )

    def to_dict(self, include_dunders: bool = False) -> Dict[str, str]:
        """ Returns a formatted dictionary view . """
        if include_dunders:
            return {k: v for k, v in self.items()}
        return {k: v for k, v in self.items() if not k.startswith("_")}

    def thats_all(self, border_char="-") -> str:
        return {k: v for k, v in vars(self).items()}

    def pretty_dict(self, border_char="-") -> str:
        """ Returns a pretty version of dictionary.

            - border_char = string used for border

            e.g.
                print(self.pretty(border_char = "˚`†´"))
        """
        return self.table_list(
            [f"{k:<20.20}: {v}" for k, v in self.to_dict().items()],
            border_char=border_char,
        )

    def to_json(self, sort_keys=True, indent=2):
        return json.dumps(
            version.to_dict(), indent=indent, sort_keys=sort_keys
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pathlib import Path
from dataclasses import dataclass, field, Field
from os import linesep as NL

vars_width = len(max(vars(Path))) + 2


@dataclass
class VarsList:
    obj: Field = field()

#     def __str__(self):
#         return str(object())


v = VarsList(Path)
print(v)
print(v.__dict__.items())

print(vars_width)
vars_list = NL.join(
    [f"{k:<{vars_width}.{vars_width}}:  {v}" for k, v in vars(Path).items()])
print(vars_list)



# -*- coding: utf-8 -*-
# 'Standard Library'
import os

with open("README.md", "w") as mkdn:
    mkdn.write("Screenshots\n===\n")
    for f in [f for f in os.listdir(".") if os.path.isfile(f)]:
        mkdn.write("`%s`\n\n" % f)
        mkdn.write("![image](%s)\n\n" % f)



#!/usr/bin/env python
# grep_some_condition.py
# https://stackoverflow.com/a/5463419

# import fileinput
from typing import Dict, List


def protected_exec(code: str):
    """ Restrict access to only ['__builtins__'] for code execution. """
    return exec(code, {})


def limited_exec(code: str, global_dict: Dict = {}):
    """ Restrict access to only `global_dict` for code execution.

        The default global_dict of {} restricts access to only ['__builtins__'] for code execution.

        e.g.
        limited_exec("print(factorial(5))", {"factorial": factorial})

        # https://www.geeksforgeeks.org/exec-in-python/
        """
    return exec(code, global_dict)


def remove_lines(test_condition='line.startswith("#TT")',
                 inplace=True,
                 backup=".bak"):
    for line in fileinput.input(inplace=inplace, backup=backup):
        if line.startswith("#TT"):
            print(line),  # this goes to the current file


def remove_inplace(filenames: List[str]):
    try:
        import in_place
    except ImportError as e:
        raise ImportError("module <in_place> was not imported correctly", e)
    else:
        with in_place.InPlace(filenames) as file:
            for line in file:
                line = line.replace("test", "testZ")
                file.write(line)


remove_inplace('line.startswith("#ST")')
""" alternative to `fileinput`

    fileinput module has very ugly API, I find beautiful module for this task - in_place, example for Python 3:

    import in_place

    with in_place.InPlace('data.txt') as file:
        for line in file:
            line = line.replace('test', 'testZ')
            file.write(line)
    main difference from fileinput:

    Instead of hijacking sys.stdout, a new filehandle is returned for writing.
    The filehandle supports all of the standard I/O methods, not just readline().
    """



# 'Standard Library'
import sys

from contextlib import closing

# 'package imports'
# from webpage import *
import webpage

# 'third party'
import requests

from bs4 import BeautifulSoup
from requests import Response, get
from requests.exceptions import RequestException

from typing import List

_debug_: bool = True


def get_mathematicians():
    # example use case
    url = "http://www.fabpedigree.com/james/mathmen.htm"
    w = WebSite(url)
    print(w)
    html = BeautifulSoup(WebSite(url), "html.parser")

    for i, li in enumerate(html.select("li")):
        print(i, li.text)


def _test_(args):
    for url in args:
        w = webpage.WebPage(url)
        w = WebSite(url)
        r = w.resp
        print(w)
        print(r)
        print(w.is_html(r))
        print(w.encoding)
        print(w.resp.encoding)
        print(w.resp.apparent_encoding)
        print()
        get_mathematicians()

        content_type = w.resp.headers["Content-Type"].lower()
        print(content_type)
        print(w.resp.headers["Content-Type"].lower())


def _main_(args: List):
    """
    CLI script main entry point.
    """
    url = "https://realpython.com/blog/"
    if _debug_:
        args.extend(url)
        _test_(args)


if __name__ == "__main__":  # if script is loaded directly from CLI
    _main_(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" # TODO -- @update `AutoSys` package
        copyright (c) 2018 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# 'Standard Library'
import string

from os import linesep
from random import choice

from typing import List

NUL: str = chr(0)
NL: str = linesep

STR_ALPHA: str = string.ascii_letters
STR_ALPHANUMERIC: str = string.ascii_letters + string.digits
STR_HEX: str = string.hexdigits
STR_NAMES: str = string.ascii_letters + " _-"
STR_WHITESPACE: str = string.whitespace
STR_PRINTABLE: str = string.printable
STR_PUNCTUATION: str = string.punctuation


def random_string(string_length=8, pool: str = STR_NAMES):
    """ Returns a string of length 'string_length' made up of characters
        from the string 'pool' """
    return "".join(choice(pool) for _ in range(string_length))


__all__: List[str] = [
    "NL",
    "NUL",
    "STR_ALPHA",
    "STR_ALPHANUMERIC",
    "STR_HEX",
    "STR_NAMES",
    "STR_PRINTABLE",
    "STR_PUNCTUATION",
    "STR_WHITESPACE",
    "random_string",
    "string",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" as_80col.py - cli utility with string and color functions. """
# copyright (c) 2019 Michael Treanor
# https://www.github.com/skeptycal
# https://www.twitter.com/skeptycal

# 'Standard Library'
import os
import textwrap

# 'package imports'
from autosys.as_constants import PY_ENV
# import autosys.as_constants
from autosys.as_system import py_shell
from autosys.colors import COLORS, STYLES, color

# sys.path.insert(0, os.path.abspath("."))

# Turn off color when using ipython, otherwise leave it on
ENABLE_COLOR: bool = True
if py_shell() in ["ipython", "ipython-notebook"]:
    ENABLE_COLOR = False


def test_styles(bg, fg):
    for style in (None,) + STYLES:
        cname = fg or "default"
        # renamed None color to default to avoid confusion wiht normal style
        if cname.startswith("bright"):
            cname = cname[6:].upper()
        text = cname[:5].ljust(6)
        print(color(text, fg=fg, bg=bg, style=style), end=" ")
    print()


# doubled number of colors, so have to split test into halves to
# test on standard 80-column terminal

colors = [None] + list(COLORS[:8])
brights = list(COLORS[8:])

for bg in colors:
    for fg in colors:
        test_styles(bg, fg)
    for fg in brights:
        test_styles(bg, fg)
for bg in brights:
    for fg in colors:
        test_styles(bg, fg)
    for fg in brights:
        test_styles(bg, fg)

for i in range(256):
    if i % 64 == 0:
        print()
    print(color(" ", bg=i), end="")

print()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# 'Standard Library'
import re

regex1 = r"\A[\w\s:\-<>\(\)\+\*]+\Z"
regex2 = r"\A[\w\s:\-<>\(\)\+\*\&\@]+\Z"
regex3 = r"[^\w\s\-\(\)\>\^\<\&\@\!\:\+\=\*]"

REGEX_SAFE_STRING = r"[^\w\s\-\(\)\>\^\<\&\@\!\:\+\=\*\'\"]"
RE_SAFE_STRING = re.compile(REGEX_SAFE_STRING, re.IGNORECASE | re.MULTILINE)

# https://regex101.com/r/9YvEas/1

RE_FLAGS_DEBUG = re.IGNORECASE | re.MULTILINE | re.VERBOSE | re.DEBUG

test_str = ("def fibonacci_2(n: int) -> int:\n" +
            "    if n > 1:\n" +
            "        return fibonacci(n - 1) + fibonacci(n - 2)\n" +
            "    else:\n" +
            "        return n=^&*(@@!)")

matches = re.search(string=test_str, pattern=RE_SAFE_STRING)


def is_safe(s) -> (bool):
    return re.search(string=s, pattern=RE_SAFE_STRING) == True


def return_matches(s, joined=True):
    tmp = [x for x in s if x in REGEX_SAFE_STRING]
    if joined:
        return ''.join(tmp)
    return tmp


def _mat(s, p, x):
    if s in p:
        return x
    return s


def remove_matches(s, joined=True):
    tmp = [_mat(x, REGEX_SAFE_STRING, '')
           for x in s.strip()]
    if joined:
        return ''.join(tmp)
    return tmp


def replace_matches(s, joined=True):
    tmp = ''.join([_mat(x, REGEX_SAFE_STRING, '_')
                   for x in s.strip()]).splitlines()
    if joined:
        return '\n'.join(tmp).strip(' _')
    return tmp


print(test_str)
print('---------------------------')

print(return_matches(test_str))
print('---------------------------')
print(remove_matches(test_str))
print('---------------------------')
print(replace_matches(test_str))
# print(is_safe(test_str))



"""
Ref: https://stackoverflow.com/a/6147060/9878098
There is a specific example of when you should use del (there may be others, but I know about this one off hand) when you are using sys.exc_info() to inspect an exception. This function returns a tuple, the type of exception that was raised, the message, and a traceback.

The first two values are usually sufficient to diagnose an error and act on it, but the third contains the entire call stack between where the exception was raised and where the the exception is caught. In particular, if you do something like

```
try:
    do_evil()
except:
    exc_type, exc_value, tb = sys.exc_info()
    if something(exc_value):
        raise
```

the traceback, tb ends up in the locals of the call stack, creating a circular reference that cannot be garbage collected. Thus, it is important to do:
"""

try:
    do_evil()
except:
    exc_type, exc_value, tb = sys.exc_info()
    del tb
    if something(exc_value):
        raise

"""
to break the circular reference. In many cases where you would want to call sys.exc_info(), like with metaclass magic, the traceback is useful, so you have to make sure that you clean it up before you can possibly leave the exception handler. If you don't need the traceback, you should delete it immediately, or just do:

```
exc_type, exc_value = sys.exc_info()[:2]
```

To avoid it all together.
"""




import sys
import os
import mimetypes
from os import linesep as NL
from sys import stderr, stdout
from pathlib import Path
from wsgiref import simple_server, util

try:
    DEFAULT_ENCODING
except:
    from locale import getpreferredencoding
    DEFAULT_ENCODING: str = getpreferredencoding(do_setlocale=True) or "utf-8"


def get_html_template() -> (str):
    return """<!DOCTYPE html>
<html lang="en">
    <head>
        <title>Learning the DOM</title>
    </head>

    <body>
        <h1>Document Object Model</h1>
        {}
    </body>
</html>
"""


def get_html_file() -> (str):
    template = get_html_template()
    body = ''
    body_list = []
    for f in Path().iterdir():
        body_list.append(str(f))
    body = NL.join(body_list)
    return template.format(body)


def create_index_file(file_name='index.html'):
    with open(file_name, mode='w') as fh:
        fh.write(get_html_file())


data = get_html_file()
print(data)
create_index_file()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2012 Matt Martz
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# 'future imports'
from __future__ import absolute_import

# 'Standard Library'
import csv
import datetime
import errno
import math
import os
import platform
import re
import signal
import socket
import sys
import threading
import timeit
import xml.parsers.expat

try:
    import gzip
    GZIP_BASE = gzip.GzipFile
except ImportError:
    gzip = None
    GZIP_BASE = object

__version__ = '2.1.2'


class FakeShutdownEvent(object):
    """Class to fake a threading.Event.isSet so that users of this module
    are not required to register their own threading.Event()
    """
    @staticmethod
    def isSet():
        "Dummy method to always return false" ""
        return False


# Some global variables we use
DEBUG = False
_GLOBAL_DEFAULT_TIMEOUT = object()
PY25PLUS = sys.version_info[:2] >= (2, 5)
PY26PLUS = sys.version_info[:2] >= (2, 6)
PY32PLUS = sys.version_info[:2] >= (3, 2)

# Begin import game to handle Python 2 and Python 3
try:
    import json
except ImportError:
    try:
        import simplejson as json
    except ImportError:
        json = None

try:
    import xml.etree.ElementTree as ET
    try:
        from xml.etree.ElementTree import _Element as ET_Element
    except ImportError:
        pass
except ImportError:
    from xml.dom import minidom as DOM
    from xml.parsers.expat import ExpatError
    ET = None

try:
    from urllib2 import (urlopen, Request, HTTPError, URLError,
                         AbstractHTTPHandler, ProxyHandler,
                         HTTPDefaultErrorHandler, HTTPRedirectHandler,
                         HTTPErrorProcessor, OpenerDirector)
except ImportError:
    from urllib.request import (urlopen, Request, HTTPError, URLError,
                                AbstractHTTPHandler, ProxyHandler,
                                HTTPDefaultErrorHandler, HTTPRedirectHandler,
                                HTTPErrorProcessor, OpenerDirector)

try:
    from httplib import HTTPConnection, BadStatusLine
except ImportError:
    from http.client import HTTPConnection, BadStatusLine

try:
    from httplib import HTTPSConnection
except ImportError:
    try:
        from http.client import HTTPSConnection
    except ImportError:
        HTTPSConnection = None

try:
    from httplib import FakeSocket
except ImportError:
    FakeSocket = None

try:
    from Queue import Queue
except ImportError:
    from queue import Queue

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

try:
    from urlparse import parse_qs
except ImportError:
    try:
        from urllib.parse import parse_qs
    except ImportError:
        from cgi import parse_qs

try:
    from hashlib import md5
except ImportError:
    from md5 import md5

try:
    from argparse import ArgumentParser as ArgParser
    from argparse import SUPPRESS as ARG_SUPPRESS
    PARSER_TYPE_INT = int
    PARSER_TYPE_STR = str
    PARSER_TYPE_FLOAT = float
except ImportError:
    from optparse import OptionParser as ArgParser
    from optparse import SUPPRESS_HELP as ARG_SUPPRESS
    PARSER_TYPE_INT = 'int'
    PARSER_TYPE_STR = 'string'
    PARSER_TYPE_FLOAT = 'float'

try:
    from cStringIO import StringIO
    BytesIO = None
except ImportError:
    try:
        from StringIO import StringIO
        BytesIO = None
    except ImportError:
        from io import StringIO, BytesIO

try:
    import __builtin__
except ImportError:
    import builtins
    from io import TextIOWrapper, FileIO

    class _Py3Utf8Output(TextIOWrapper):
        """UTF-8 encoded wrapper around stdout for py3, to override
        ASCII stdout
        """
        def __init__(self, f, **kwargs):
            buf = FileIO(f.fileno(), 'w')
            super(_Py3Utf8Output, self).__init__(buf,
                                                 encoding='utf8',
                                                 errors='strict')

        def write(self, s):
            super(_Py3Utf8Output, self).write(s)
            self.flush()

    _py3_print = getattr(builtins, 'print')
    try:
        _py3_utf8_stdout = _Py3Utf8Output(sys.stdout)
        _py3_utf8_stderr = _Py3Utf8Output(sys.stderr)
    except OSError:
        # sys.stdout/sys.stderr is not a compatible stdout/stderr object
        # just use it and hope things go ok
        _py3_utf8_stdout = sys.stdout
        _py3_utf8_stderr = sys.stderr

    def to_utf8(v):
        """No-op encode to utf-8 for py3"""
        return v

    def print_(*args, **kwargs):
        """Wrapper function for py3 to print, with a utf-8 encoded stdout"""
        if kwargs.get('file') == sys.stderr:
            kwargs['file'] = _py3_utf8_stderr
        else:
            kwargs['file'] = kwargs.get('file', _py3_utf8_stdout)
        _py3_print(*args, **kwargs)
else:
    del __builtin__

    def to_utf8(v):
        """Encode value to utf-8 if possible for py2"""
        try:
            return v.encode('utf8', 'strict')
        except AttributeError:
            return v

    def print_(*args, **kwargs):
        """The new-style print function for Python 2.4 and 2.5.

        Taken from https://pypi.python.org/pypi/six/

        Modified to set encoding to UTF-8 always, and to flush after write
        """
        fp = kwargs.pop("file", sys.stdout)
        if fp is None:
            return

        def write(data):
            if not isinstance(data, basestring):
                data = str(data)
            # If the file has an encoding, encode unicode with it.
            encoding = 'utf8'  # Always trust UTF-8 for output
            if (isinstance(fp, file) and isinstance(data, unicode)
                    and encoding is not None):
                errors = getattr(fp, "errors", None)
                if errors is None:
                    errors = "strict"
                data = data.encode(encoding, errors)
            fp.write(data)
            fp.flush()

        want_unicode = False
        sep = kwargs.pop("sep", None)
        if sep is not None:
            if isinstance(sep, unicode):
                want_unicode = True
            elif not isinstance(sep, str):
                raise TypeError("sep must be None or a string")
        end = kwargs.pop("end", None)
        if end is not None:
            if isinstance(end, unicode):
                want_unicode = True
            elif not isinstance(end, str):
                raise TypeError("end must be None or a string")
        if kwargs:
            raise TypeError("invalid keyword arguments to print()")
        if not want_unicode:
            for arg in args:
                if isinstance(arg, unicode):
                    want_unicode = True
                    break
        if want_unicode:
            newline = unicode("\n")
            space = unicode(" ")
        else:
            newline = "\n"
            space = " "
        if sep is None:
            sep = space
        if end is None:
            end = newline
        for i, arg in enumerate(args):
            if i:
                write(sep)
            write(arg)
        write(end)


if PY32PLUS:
    etree_iter = ET.Element.iter
elif PY25PLUS:
    etree_iter = ET_Element.getiterator

if PY26PLUS:
    thread_is_alive = threading.Thread.is_alive
else:
    thread_is_alive = threading.Thread.isAlive

# Exception "constants" to support Python 2 through Python 3
try:
    import ssl
    try:
        CERT_ERROR = (ssl.CertificateError, )
    except AttributeError:
        CERT_ERROR = tuple()

    HTTP_ERRORS = (
        (HTTPError, URLError, socket.error, ssl.SSLError, BadStatusLine) +
        CERT_ERROR)
except ImportError:
    ssl = None
    HTTP_ERRORS = (HTTPError, URLError, socket.error, BadStatusLine)


class SpeedtestException(Exception):
    """Base exception for this module"""


class SpeedtestCLIError(SpeedtestException):
    """Generic exception for raising errors during CLI operation"""


class SpeedtestHTTPError(SpeedtestException):
    """Base HTTP exception for this module"""


class SpeedtestConfigError(SpeedtestException):
    """Configuration XML is invalid"""


class SpeedtestServersError(SpeedtestException):
    """Servers XML is invalid"""


class ConfigRetrievalError(SpeedtestHTTPError):
    """Could not retrieve config.php"""


class ServersRetrievalError(SpeedtestHTTPError):
    """Could not retrieve speedtest-servers.php"""


class InvalidServerIDType(SpeedtestException):
    """Server ID used for filtering was not an integer"""


class NoMatchedServers(SpeedtestException):
    """No servers matched when filtering"""


class SpeedtestMiniConnectFailure(SpeedtestException):
    """Could not connect to the provided speedtest mini server"""


class InvalidSpeedtestMiniServer(SpeedtestException):
    """Server provided as a speedtest mini server does not actually appear
    to be a speedtest mini server
    """


class ShareResultsConnectFailure(SpeedtestException):
    """Could not connect to speedtest.net API to POST results"""


class ShareResultsSubmitFailure(SpeedtestException):
    """Unable to successfully POST results to speedtest.net API after
    connection
    """


class SpeedtestUploadTimeout(SpeedtestException):
    """testlength configuration reached during upload
    Used to ensure the upload halts when no additional data should be sent
    """


class SpeedtestBestServerFailure(SpeedtestException):
    """Unable to determine best server"""


class SpeedtestMissingBestServer(SpeedtestException):
    """get_best_server not called or not able to determine best server"""


def create_connection(address,
                      timeout=_GLOBAL_DEFAULT_TIMEOUT,
                      source_address=None):
    """Connect to *address* and return the socket object.

    Convenience function.  Connect to *address* (a 2-tuple ``(host,
    port)``) and return the socket object.  Passing the optional
    *timeout* parameter will set the timeout on the socket instance
    before attempting to connect.  If no *timeout* is supplied, the
    global default timeout setting returned by :func:`getdefaulttimeout`
    is used.  If *source_address* is set it must be a tuple of (host, port)
    for the socket to bind as a source address before making the connection.
    An host of '' or port 0 tells the OS to use the default.

    Largely vendored from Python 2.7, modified to work with Python 2.4
    """

    host, port = address
    err = None
    for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        sock = None
        try:
            sock = socket.socket(af, socktype, proto)
            if timeout is not _GLOBAL_DEFAULT_TIMEOUT:
                sock.settimeout(float(timeout))
            if source_address:
                sock.bind(source_address)
            sock.connect(sa)
            return sock

        except socket.error:
            err = get_exception()
            if sock is not None:
                sock.close()

    if err is not None:
        raise err
    else:
        raise socket.error("getaddrinfo returns an empty list")


class SpeedtestHTTPConnection(HTTPConnection):
    """Custom HTTPConnection to support source_address across
    Python 2.4 - Python 3
    """
    def __init__(self, *args, **kwargs):
        source_address = kwargs.pop('source_address', None)
        timeout = kwargs.pop('timeout', 10)

        self._tunnel_host = None

        HTTPConnection.__init__(self, *args, **kwargs)

        self.source_address = source_address
        self.timeout = timeout

    def connect(self):
        """Connect to the host and port specified in __init__."""
        try:
            self.sock = socket.create_connection(
                (self.host, self.port), self.timeout, self.source_address)
        except (AttributeError, TypeError):
            self.sock = create_connection((self.host, self.port), self.timeout,
                                          self.source_address)

        if self._tunnel_host:
            self._tunnel()


if HTTPSConnection:

    class SpeedtestHTTPSConnection(HTTPSConnection):
        """Custom HTTPSConnection to support source_address across
        Python 2.4 - Python 3
        """
        default_port = 443

        def __init__(self, *args, **kwargs):
            source_address = kwargs.pop('source_address', None)
            timeout = kwargs.pop('timeout', 10)

            self._tunnel_host = None

            HTTPSConnection.__init__(self, *args, **kwargs)

            self.timeout = timeout
            self.source_address = source_address

        def connect(self):
            "Connect to a host on a given (SSL) port."
            try:
                self.sock = socket.create_connection(
                    (self.host, self.port), self.timeout, self.source_address)
            except (AttributeError, TypeError):
                self.sock = create_connection(
                    (self.host, self.port), self.timeout, self.source_address)

            if self._tunnel_host:
                self._tunnel()

            if ssl:
                try:
                    kwargs = {}
                    if hasattr(ssl, 'SSLContext'):
                        if self._tunnel_host:
                            kwargs['server_hostname'] = self._tunnel_host
                        else:
                            kwargs['server_hostname'] = self.host
                    self.sock = self._context.wrap_socket(self.sock, **kwargs)
                except AttributeError:
                    self.sock = ssl.wrap_socket(self.sock)
                    try:
                        self.sock.server_hostname = self.host
                    except AttributeError:
                        pass
            elif FakeSocket:
                # Python 2.4/2.5 support
                try:
                    self.sock = FakeSocket(self.sock, socket.ssl(self.sock))
                except AttributeError:
                    raise SpeedtestException(
                        'This version of Python does not support HTTPS/SSL '
                        'functionality')
            else:
                raise SpeedtestException(
                    'This version of Python does not support HTTPS/SSL '
                    'functionality')


def _build_connection(connection, source_address, timeout, context=None):
    """Cross Python 2.4 - Python 3 callable to build an ``HTTPConnection`` or
    ``HTTPSConnection`` with the args we need

    Called from ``http(s)_open`` methods of ``SpeedtestHTTPHandler`` or
    ``SpeedtestHTTPSHandler``
    """
    def inner(host, **kwargs):
        kwargs.update({'source_address': source_address, 'timeout': timeout})
        if context:
            kwargs['context'] = context
        return connection(host, **kwargs)

    return inner


class SpeedtestHTTPHandler(AbstractHTTPHandler):
    """Custom ``HTTPHandler`` that can build a ``HTTPConnection`` with the
    args we need for ``source_address`` and ``timeout``
    """
    def __init__(self, debuglevel=0, source_address=None, timeout=10):
        AbstractHTTPHandler.__init__(self, debuglevel)
        self.source_address = source_address
        self.timeout = timeout

    def http_open(self, req):
        return self.do_open(
            _build_connection(SpeedtestHTTPConnection, self.source_address,
                              self.timeout), req)

    http_request = AbstractHTTPHandler.do_request_


class SpeedtestHTTPSHandler(AbstractHTTPHandler):
    """Custom ``HTTPSHandler`` that can build a ``HTTPSConnection`` with the
    args we need for ``source_address`` and ``timeout``
    """
    def __init__(self,
                 debuglevel=0,
                 context=None,
                 source_address=None,
                 timeout=10):
        AbstractHTTPHandler.__init__(self, debuglevel)
        self._context = context
        self.source_address = source_address
        self.timeout = timeout

    def https_open(self, req):
        return self.do_open(
            _build_connection(
                SpeedtestHTTPSConnection,
                self.source_address,
                self.timeout,
                context=self._context,
            ), req)

    https_request = AbstractHTTPHandler.do_request_


def build_opener(source_address=None, timeout=10):
    """Function similar to ``urllib2.build_opener`` that will build
    an ``OpenerDirector`` with the explicit handlers we want,
    ``source_address`` for binding, ``timeout`` and our custom
    `User-Agent`
    """

    printer('Timeout set to %d' % timeout, debug=True)

    if source_address:
        source_address_tuple = (source_address, 0)
        printer('Binding to source address: %r' % (source_address_tuple, ),
                debug=True)
    else:
        source_address_tuple = None

    handlers = [
        ProxyHandler(),
        SpeedtestHTTPHandler(source_address=source_address_tuple,
                             timeout=timeout),
        SpeedtestHTTPSHandler(source_address=source_address_tuple,
                              timeout=timeout),
        HTTPDefaultErrorHandler(),
        HTTPRedirectHandler(),
        HTTPErrorProcessor()
    ]

    opener = OpenerDirector()
    opener.addheaders = [('User-agent', build_user_agent())]

    for handler in handlers:
        opener.add_handler(handler)

    return opener


class GzipDecodedResponse(GZIP_BASE):
    """A file-like object to decode a response encoded with the gzip
    method, as described in RFC 1952.

    Largely copied from ``xmlrpclib``/``xmlrpc.client`` and modified
    to work for py2.4-py3
    """
    def __init__(self, response):
        # response doesn't support tell() and read(), required by
        # GzipFile
        if not gzip:
            raise SpeedtestHTTPError('HTTP response body is gzip encoded, '
                                     'but gzip support is not available')
        IO = BytesIO or StringIO
        self.io = IO()
        while 1:
            chunk = response.read(1024)
            if len(chunk) == 0:
                break
            self.io.write(chunk)
        self.io.seek(0)
        gzip.GzipFile.__init__(self, mode='rb', fileobj=self.io)

    def close(self):
        try:
            gzip.GzipFile.close(self)
        finally:
            self.io.close()


def get_exception():
    """Helper function to work with py2.4-py3 for getting the current
    exception in a try/except block
    """
    return sys.exc_info()[1]


def distance(origin, destination):
    """Determine distance between 2 sets of [lat,lon] in km"""

    lat1, lon1 = origin
    lat2, lon2 = destination
    radius = 6371  # km

    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    a = (math.sin(dlat / 2) * math.sin(dlat / 2) +
         math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) *
         math.sin(dlon / 2) * math.sin(dlon / 2))
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    d = radius * c

    return d


def build_user_agent():
    """Build a Mozilla/5.0 compatible User-Agent string"""

    ua_tuple = ('Mozilla/5.0', '(%s; U; %s; en-us)' %
                (platform.platform(), platform.architecture()[0]),
                'Python/%s' % platform.python_version(), '(KHTML, like Gecko)',
                'speedtest-cli/%s' % __version__)
    user_agent = ' '.join(ua_tuple)
    printer('User-Agent: %s' % user_agent, debug=True)
    return user_agent


def build_request(url, data=None, headers=None, bump='0', secure=False):
    """Build a urllib2 request object

    This function automatically adds a User-Agent header to all requests

    """

    if not headers:
        headers = {}

    if url[0] == ':':
        scheme = ('http', 'https')[bool(secure)]
        schemed_url = '%s%s' % (scheme, url)
    else:
        schemed_url = url

    if '?' in url:
        delim = '&'
    else:
        delim = '?'

    # WHO YOU GONNA CALL? CACHE BUSTERS!
    final_url = '%s%sx=%s.%s' % (schemed_url, delim,
                                 int(timeit.time.time() * 1000), bump)

    headers.update({
        'Cache-Control': 'no-cache',
    })

    printer('%s %s' % (('GET', 'POST')[bool(data)], final_url), debug=True)

    return Request(final_url, data=data, headers=headers)


def catch_request(request, opener=None):
    """Helper function to catch common exceptions encountered when
    establishing a connection with a HTTP/HTTPS request

    """

    if opener:
        _open = opener.open
    else:
        _open = urlopen

    try:
        uh = _open(request)
        if request.get_full_url() != uh.geturl():
            printer('Redirected to %s' % uh.geturl(), debug=True)
        return uh, False
    except HTTP_ERRORS:
        e = get_exception()
        return None, e


def get_response_stream(response):
    """Helper function to return either a Gzip reader if
    ``Content-Encoding`` is ``gzip`` otherwise the response itself

    """

    try:
        getheader = response.headers.getheader
    except AttributeError:
        getheader = response.getheader

    if getheader('content-encoding') == 'gzip':
        return GzipDecodedResponse(response)

    return response


def get_attributes_by_tag_name(dom, tag_name):
    """Retrieve an attribute from an XML document and return it in a
    consistent format

    Only used with xml.dom.minidom, which is likely only to be used
    with python versions older than 2.5
    """
    elem = dom.getElementsByTagName(tag_name)[0]
    return dict(list(elem.attributes.items()))


def print_dots(shutdown_event):
    """Built in callback function used by Thread classes for printing
    status
    """
    def inner(current, total, start=False, end=False):
        if shutdown_event.isSet():
            return

        sys.stdout.write('.')
        if current + 1 == total and end is True:
            sys.stdout.write('\n')
        sys.stdout.flush()

    return inner


def do_nothing(*args, **kwargs):
    pass


class HTTPDownloader(threading.Thread):
    """Thread class for retrieving a URL"""
    def __init__(self,
                 i,
                 request,
                 start,
                 timeout,
                 opener=None,
                 shutdown_event=None):
        threading.Thread.__init__(self)
        self.request = request
        self.result = [0]
        self.starttime = start
        self.timeout = timeout
        self.i = i
        if opener:
            self._opener = opener.open
        else:
            self._opener = urlopen

        if shutdown_event:
            self._shutdown_event = shutdown_event
        else:
            self._shutdown_event = FakeShutdownEvent()

    def run(self):
        try:
            if (timeit.default_timer() - self.starttime) <= self.timeout:
                f = self._opener(self.request)
                while (
                        not self._shutdown_event.isSet() and
                    (timeit.default_timer() - self.starttime) <= self.timeout):
                    self.result.append(len(f.read(10240)))
                    if self.result[-1] == 0:
                        break
                f.close()
        except IOError:
            pass


class HTTPUploaderData(object):
    """File like object to improve cutting off the upload once the timeout
    has been reached
    """
    def __init__(self, length, start, timeout, shutdown_event=None):
        self.length = length
        self.start = start
        self.timeout = timeout

        if shutdown_event:
            self._shutdown_event = shutdown_event
        else:
            self._shutdown_event = FakeShutdownEvent()

        self._data = None

        self.total = [0]

    def pre_allocate(self):
        chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        multiplier = int(round(int(self.length) / 36.0))
        IO = BytesIO or StringIO
        try:
            self._data = IO(
                ('content1=%s' %
                 (chars * multiplier)[0:int(self.length) - 9]).encode())
        except MemoryError:
            raise SpeedtestCLIError(
                'Insufficient memory to pre-allocate upload data. Please '
                'use --no-pre-allocate')

    @property
    def data(self):
        if not self._data:
            self.pre_allocate()
        return self._data

    def read(self, n=10240):
        if ((timeit.default_timer() - self.start) <= self.timeout
                and not self._shutdown_event.isSet()):
            chunk = self.data.read(n)
            self.total.append(len(chunk))
            return chunk
        else:
            raise SpeedtestUploadTimeout()

    def __len__(self):
        return self.length


class HTTPUploader(threading.Thread):
    """Thread class for putting a URL"""
    def __init__(self,
                 i,
                 request,
                 start,
                 size,
                 timeout,
                 opener=None,
                 shutdown_event=None):
        threading.Thread.__init__(self)
        self.request = request
        self.request.data.start = self.starttime = start
        self.size = size
        self.result = None
        self.timeout = timeout
        self.i = i

        if opener:
            self._opener = opener.open
        else:
            self._opener = urlopen

        if shutdown_event:
            self._shutdown_event = shutdown_event
        else:
            self._shutdown_event = FakeShutdownEvent()

    def run(self):
        request = self.request
        try:
            if ((timeit.default_timer() - self.starttime) <= self.timeout
                    and not self._shutdown_event.isSet()):
                try:
                    f = self._opener(request)
                except TypeError:
                    # PY24 expects a string or buffer
                    # This also causes issues with Ctrl-C, but we will concede
                    # for the moment that Ctrl-C on PY24 isn't immediate
                    request = build_request(self.request.get_full_url(),
                                            data=request.data.read(self.size))
                    f = self._opener(request)
                f.read(11)
                f.close()
                self.result = sum(self.request.data.total)
            else:
                self.result = 0
        except (IOError, SpeedtestUploadTimeout):
            self.result = sum(self.request.data.total)


class SpeedtestResults(object):
    """Class for holding the results of a speedtest, including:

    Download speed
    Upload speed
    Ping/Latency to test server
    Data about server that the test was run against

    Additionally this class can return a result data as a dictionary or CSV,
    as well as submit a POST of the result data to the speedtest.net API
    to get a share results image link.
    """
    def __init__(self,
                 download=0,
                 upload=0,
                 ping=0,
                 server=None,
                 client=None,
                 opener=None,
                 secure=False):
        self.download = download
        self.upload = upload
        self.ping = ping
        if server is None:
            self.server = {}
        else:
            self.server = server
        self.client = client or {}

        self._share = None
        self.timestamp = '%sZ' % datetime.datetime.utcnow().isoformat()
        self.bytes_received = 0
        self.bytes_sent = 0

        if opener:
            self._opener = opener
        else:
            self._opener = build_opener()

        self._secure = secure

    def __repr__(self):
        return repr(self.dict())

    def share(self):
        """POST data to the speedtest.net API to obtain a share results
        link
        """

        if self._share:
            return self._share

        download = int(round(self.download / 1000.0, 0))
        ping = int(round(self.ping, 0))
        upload = int(round(self.upload / 1000.0, 0))

        # Build the request to send results back to speedtest.net
        # We use a list instead of a dict because the API expects parameters
        # in a certain order
        api_data = [
            'recommendedserverid=%s' % self.server['id'],
            'ping=%s' % ping,
            'screenresolution=',
            'promo=',
            'download=%s' % download,
            'screendpi=',
            'upload=%s' % upload,
            'testmethod=http',
            'hash=%s' % md5(
                ('%s-%s-%s-%s' %
                 (ping, upload, download, '297aae72')).encode()).hexdigest(),
            'touchscreen=none',
            'startmode=pingselect',
            'accuracy=1',
            'bytesreceived=%s' % self.bytes_received,
            'bytessent=%s' % self.bytes_sent,
            'serverid=%s' % self.server['id'],
        ]

        headers = {'Referer': 'http://c.speedtest.net/flash/speedtest.swf'}
        request = build_request('://www.speedtest.net/api/api.php',
                                data='&'.join(api_data).encode(),
                                headers=headers,
                                secure=self._secure)
        f, e = catch_request(request, opener=self._opener)
        if e:
            raise ShareResultsConnectFailure(e)

        response = f.read()
        code = f.code
        f.close()

        if int(code) != 200:
            raise ShareResultsSubmitFailure('Could not submit results to '
                                            'speedtest.net')

        qsargs = parse_qs(response.decode())
        resultid = qsargs.get('resultid')
        if not resultid or len(resultid) != 1:
            raise ShareResultsSubmitFailure('Could not submit results to '
                                            'speedtest.net')

        self._share = 'http://www.speedtest.net/result/%s.png' % resultid[0]

        return self._share

    def dict(self):
        """Return dictionary of result data"""

        return {
            'download': self.download,
            'upload': self.upload,
            'ping': self.ping,
            'server': self.server,
            'timestamp': self.timestamp,
            'bytes_sent': self.bytes_sent,
            'bytes_received': self.bytes_received,
            'share': self._share,
            'client': self.client,
        }

    @staticmethod
    def csv_header(delimiter=','):
        """Return CSV Headers"""

        row = [
            'Server ID', 'Sponsor', 'Server Name', 'Timestamp', 'Distance',
            'Ping', 'Download', 'Upload', 'Share', 'IP Address'
        ]
        out = StringIO()
        writer = csv.writer(out, delimiter=delimiter, lineterminator='')
        writer.writerow([to_utf8(v) for v in row])
        return out.getvalue()

    def csv(self, delimiter=','):
        """Return data in CSV format"""

        data = self.dict()
        out = StringIO()
        writer = csv.writer(out, delimiter=delimiter, lineterminator='')
        row = [
            data['server']['id'], data['server']['sponsor'],
            data['server']['name'], data['timestamp'], data['server']['d'],
            data['ping'], data['download'], data['upload'], self._share or '',
            self.client['ip']
        ]
        writer.writerow([to_utf8(v) for v in row])
        return out.getvalue()

    def json(self, pretty=False):
        """Return data in JSON format"""

        kwargs = {}
        if pretty:
            kwargs.update({'indent': 4, 'sort_keys': True})
        return json.dumps(self.dict(), **kwargs)


class Speedtest(object):
    """Class for performing standard speedtest.net testing operations"""
    def __init__(self,
                 config=None,
                 source_address=None,
                 timeout=10,
                 secure=False,
                 shutdown_event=None):
        self.config = {}

        self._source_address = source_address
        self._timeout = timeout
        self._opener = build_opener(source_address, timeout)

        self._secure = secure

        if shutdown_event:
            self._shutdown_event = shutdown_event
        else:
            self._shutdown_event = FakeShutdownEvent()

        self.get_config()
        if config is not None:
            self.config.update(config)

        self.servers = {}
        self.closest = []
        self._best = {}

        self.results = SpeedtestResults(
            client=self.config['client'],
            opener=self._opener,
            secure=secure,
        )

    @property
    def best(self):
        if not self._best:
            self.get_best_server()
        return self._best

    def get_config(self):
        """Download the speedtest.net configuration and return only the data
        we are interested in
        """

        headers = {}
        if gzip:
            headers['Accept-Encoding'] = 'gzip'
        request = build_request('://www.speedtest.net/speedtest-config.php',
                                headers=headers,
                                secure=self._secure)
        uh, e = catch_request(request, opener=self._opener)
        if e:
            raise ConfigRetrievalError(e)
        configxml_list = []

        stream = get_response_stream(uh)

        while 1:
            try:
                configxml_list.append(stream.read(1024))
            except (OSError, EOFError):
                raise ConfigRetrievalError(get_exception())
            if len(configxml_list[-1]) == 0:
                break
        stream.close()
        uh.close()

        if int(uh.code) != 200:
            return None

        configxml = ''.encode().join(configxml_list)

        printer('Config XML:\n%s' % configxml, debug=True)

        try:
            try:
                root = ET.fromstring(configxml)
            except ET.ParseError:
                e = get_exception()
                raise SpeedtestConfigError(
                    'Malformed speedtest.net configuration: %s' % e)
            server_config = root.find('server-config').attrib
            download = root.find('download').attrib
            upload = root.find('upload').attrib
            # times = root.find('times').attrib
            client = root.find('client').attrib

        except AttributeError:
            try:
                root = DOM.parseString(configxml)
            except ExpatError:
                e = get_exception()
                raise SpeedtestConfigError(
                    'Malformed speedtest.net configuration: %s' % e)
            server_config = get_attributes_by_tag_name(root, 'server-config')
            download = get_attributes_by_tag_name(root, 'download')
            upload = get_attributes_by_tag_name(root, 'upload')
            # times = get_attributes_by_tag_name(root, 'times')
            client = get_attributes_by_tag_name(root, 'client')

        ignore_servers = list(map(int, server_config['ignoreids'].split(',')))

        ratio = int(upload['ratio'])
        upload_max = int(upload['maxchunkcount'])
        up_sizes = [32768, 65536, 131072, 262144, 524288, 1048576, 7340032]
        sizes = {
            'upload': up_sizes[ratio - 1:],
            'download':
            [350, 500, 750, 1000, 1500, 2000, 2500, 3000, 3500, 4000]
        }

        size_count = len(sizes['upload'])

        upload_count = int(math.ceil(upload_max / size_count))

        counts = {
            'upload': upload_count,
            'download': int(download['threadsperurl'])
        }

        threads = {
            'upload': int(upload['threads']),
            'download': int(server_config['threadcount']) * 2
        }

        length = {
            'upload': int(upload['testlength']),
            'download': int(download['testlength'])
        }

        self.config.update({
            'client': client,
            'ignore_servers': ignore_servers,
            'sizes': sizes,
            'counts': counts,
            'threads': threads,
            'length': length,
            'upload_max': upload_count * size_count
        })

        try:
            self.lat_lon = (float(client['lat']), float(client['lon']))
        except ValueError:
            raise SpeedtestConfigError('Unknown location: lat=%r lon=%r' %
                                       (client.get('lat'), client.get('lon')))

        printer('Config:\n%r' % self.config, debug=True)

        return self.config

    def get_servers(self, servers=None, exclude=None):
        """Retrieve a the list of speedtest.net servers, optionally filtered
        to servers matching those specified in the ``servers`` argument
        """
        if servers is None:
            servers = []

        if exclude is None:
            exclude = []

        self.servers.clear()

        for server_list in (servers, exclude):
            for i, s in enumerate(server_list):
                try:
                    server_list[i] = int(s)
                except ValueError:
                    raise InvalidServerIDType(
                        '%s is an invalid server type, must be int' % s)

        urls = [
            '://www.speedtest.net/speedtest-servers-static.php',
            'http://c.speedtest.net/speedtest-servers-static.php',
            '://www.speedtest.net/speedtest-servers.php',
            'http://c.speedtest.net/speedtest-servers.php',
        ]

        headers = {}
        if gzip:
            headers['Accept-Encoding'] = 'gzip'

        errors = []
        for url in urls:
            try:
                request = build_request(
                    '%s?threads=%s' %
                    (url, self.config['threads']['download']),
                    headers=headers,
                    secure=self._secure)
                uh, e = catch_request(request, opener=self._opener)
                if e:
                    errors.append('%s' % e)
                    raise ServersRetrievalError()

                stream = get_response_stream(uh)

                serversxml_list = []
                while 1:
                    try:
                        serversxml_list.append(stream.read(1024))
                    except (OSError, EOFError):
                        raise ServersRetrievalError(get_exception())
                    if len(serversxml_list[-1]) == 0:
                        break

                stream.close()
                uh.close()

                if int(uh.code) != 200:
                    raise ServersRetrievalError()

                serversxml = ''.encode().join(serversxml_list)

                printer('Servers XML:\n%s' % serversxml, debug=True)

                try:
                    try:
                        try:
                            root = ET.fromstring(serversxml)
                        except ET.ParseError:
                            e = get_exception()
                            raise SpeedtestServersError(
                                'Malformed speedtest.net server list: %s' % e)
                        elements = etree_iter(root, 'server')
                    except AttributeError:
                        try:
                            root = DOM.parseString(serversxml)
                        except ExpatError:
                            e = get_exception()
                            raise SpeedtestServersError(
                                'Malformed speedtest.net server list: %s' % e)
                        elements = root.getElementsByTagName('server')
                except (SyntaxError, xml.parsers.expat.ExpatError):
                    raise ServersRetrievalError()

                for server in elements:
                    try:
                        attrib = server.attrib
                    except AttributeError:
                        attrib = dict(list(server.attributes.items()))

                    if servers and int(attrib.get('id')) not in servers:
                        continue

                    if (int(attrib.get('id')) in self.config['ignore_servers']
                            or int(attrib.get('id')) in exclude):
                        continue

                    try:
                        d = distance(self.lat_lon, (float(
                            attrib.get('lat')), float(attrib.get('lon'))))
                    except Exception:
                        continue

                    attrib['d'] = d

                    try:
                        self.servers[d].append(attrib)
                    except KeyError:
                        self.servers[d] = [attrib]

                break

            except ServersRetrievalError:
                continue

        if (servers or exclude) and not self.servers:
            raise NoMatchedServers()

        return self.servers

    def set_mini_server(self, server):
        """Instead of querying for a list of servers, set a link to a
        speedtest mini server
        """

        urlparts = urlparse(server)

        name, ext = os.path.splitext(urlparts[2])
        if ext:
            url = os.path.dirname(server)
        else:
            url = server

        request = build_request(url)
        uh, e = catch_request(request, opener=self._opener)
        if e:
            raise SpeedtestMiniConnectFailure('Failed to connect to %s' %
                                              server)
        else:
            text = uh.read()
            uh.close()

        extension = re.findall('upload_?[Ee]xtension: "([^"]+)"',
                               text.decode())
        if not extension:
            for ext in ['php', 'asp', 'aspx', 'jsp']:
                try:
                    f = self._opener.open('%s/speedtest/upload.%s' %
                                          (url, ext))
                except Exception:
                    pass
                else:
                    data = f.read().strip().decode()
                    if (f.code == 200 and len(data.splitlines()) == 1
                            and re.match('size=[0-9]', data)):
                        extension = [ext]
                        break
        if not urlparts or not extension:
            raise InvalidSpeedtestMiniServer('Invalid Speedtest Mini Server: '
                                             '%s' % server)

        self.servers = [{
            'sponsor':
            'Speedtest Mini',
            'name':
            urlparts[1],
            'd':
            0,
            'url':
            '%s/speedtest/upload.%s' % (url.rstrip('/'), extension[0]),
            'latency':
            0,
            'id':
            0
        }]

        return self.servers

    def get_closest_servers(self, limit=5):
        """Limit servers to the closest speedtest.net servers based on
        geographic distance
        """

        if not self.servers:
            self.get_servers()

        for d in sorted(self.servers.keys()):
            for s in self.servers[d]:
                self.closest.append(s)
                if len(self.closest) == limit:
                    break
            else:
                continue
            break

        printer('Closest Servers:\n%r' % self.closest, debug=True)
        return self.closest

    def get_best_server(self, servers=None):
        """Perform a speedtest.net "ping" to determine which speedtest.net
        server has the lowest latency
        """

        if not servers:
            if not self.closest:
                servers = self.get_closest_servers()
            servers = self.closest

        if self._source_address:
            source_address_tuple = (self._source_address, 0)
        else:
            source_address_tuple = None

        user_agent = build_user_agent()

        results = {}
        for server in servers:
            cum = []
            url = os.path.dirname(server['url'])
            stamp = int(timeit.time.time() * 1000)
            latency_url = '%s/latency.txt?x=%s' % (url, stamp)
            for i in range(0, 3):
                this_latency_url = '%s.%s' % (latency_url, i)
                printer('%s %s' % ('GET', this_latency_url), debug=True)
                urlparts = urlparse(latency_url)
                try:
                    if urlparts[0] == 'https':
                        h = SpeedtestHTTPSConnection(
                            urlparts[1], source_address=source_address_tuple)
                    else:
                        h = SpeedtestHTTPConnection(
                            urlparts[1], source_address=source_address_tuple)
                    headers = {'User-Agent': user_agent}
                    path = '%s?%s' % (urlparts[2], urlparts[4])
                    start = timeit.default_timer()
                    h.request("GET", path, headers=headers)
                    r = h.getresponse()
                    total = (timeit.default_timer() - start)
                except HTTP_ERRORS:
                    e = get_exception()
                    printer('ERROR: %r' % e, debug=True)
                    cum.append(3600)
                    continue

                text = r.read(9)
                if int(r.status) == 200 and text == 'test=test'.encode():
                    cum.append(total)
                else:
                    cum.append(3600)
                h.close()

            avg = round((sum(cum) / 6) * 1000.0, 3)
            results[avg] = server

        try:
            fastest = sorted(results.keys())[0]
        except IndexError:
            raise SpeedtestBestServerFailure('Unable to connect to servers to '
                                             'test latency.')
        best = results[fastest]
        best['latency'] = fastest

        self.results.ping = fastest
        self.results.server = best

        self._best.update(best)
        printer('Best Server:\n%r' % best, debug=True)
        return best

    def download(self, callback=do_nothing, threads=None):
        """Test download speed against speedtest.net

        A ``threads`` value of ``None`` will fall back to those dictated
        by the speedtest.net configuration
        """

        urls = []
        for size in self.config['sizes']['download']:
            for _ in range(0, self.config['counts']['download']):
                urls.append('%s/random%sx%s.jpg' %
                            (os.path.dirname(self.best['url']), size, size))

        request_count = len(urls)
        requests = []
        for i, url in enumerate(urls):
            requests.append(build_request(url, bump=i, secure=self._secure))

        max_threads = threads or self.config['threads']['download']
        in_flight = {'threads': 0}

        def producer(q, requests, request_count):
            for i, request in enumerate(requests):
                thread = HTTPDownloader(i,
                                        request,
                                        start,
                                        self.config['length']['download'],
                                        opener=self._opener,
                                        shutdown_event=self._shutdown_event)
                while in_flight['threads'] >= max_threads:
                    timeit.time.sleep(0.001)
                thread.start()
                q.put(thread, True)
                in_flight['threads'] += 1
                callback(i, request_count, start=True)

        finished = []

        def consumer(q, request_count):
            _is_alive = thread_is_alive
            while len(finished) < request_count:
                thread = q.get(True)
                while _is_alive(thread):
                    thread.join(timeout=0.001)
                in_flight['threads'] -= 1
                finished.append(sum(thread.result))
                callback(thread.i, request_count, end=True)

        q = Queue(max_threads)
        prod_thread = threading.Thread(target=producer,
                                       args=(q, requests, request_count))
        cons_thread = threading.Thread(target=consumer,
                                       args=(q, request_count))
        start = timeit.default_timer()
        prod_thread.start()
        cons_thread.start()
        _is_alive = thread_is_alive
        while _is_alive(prod_thread):
            prod_thread.join(timeout=0.001)
        while _is_alive(cons_thread):
            cons_thread.join(timeout=0.001)

        stop = timeit.default_timer()
        self.results.bytes_received = sum(finished)
        self.results.download = ((self.results.bytes_received /
                                  (stop - start)) * 8.0)
        if self.results.download > 100000:
            self.config['threads']['upload'] = 8
        return self.results.download

    def upload(self, callback=do_nothing, pre_allocate=True, threads=None):
        """Test upload speed against speedtest.net

        A ``threads`` value of ``None`` will fall back to those dictated
        by the speedtest.net configuration
        """

        sizes = []

        for size in self.config['sizes']['upload']:
            for _ in range(0, self.config['counts']['upload']):
                sizes.append(size)

        # request_count = len(sizes)
        request_count = self.config['upload_max']

        requests = []
        for i, size in enumerate(sizes):
            # We set ``0`` for ``start`` and handle setting the actual
            # ``start`` in ``HTTPUploader`` to get better measurements
            data = HTTPUploaderData(size,
                                    0,
                                    self.config['length']['upload'],
                                    shutdown_event=self._shutdown_event)
            if pre_allocate:
                data.pre_allocate()

            headers = {'Content-length': size}
            requests.append((build_request(self.best['url'],
                                           data,
                                           secure=self._secure,
                                           headers=headers), size))

        max_threads = threads or self.config['threads']['upload']
        in_flight = {'threads': 0}

        def producer(q, requests, request_count):
            for i, request in enumerate(requests[:request_count]):
                thread = HTTPUploader(i,
                                      request[0],
                                      start,
                                      request[1],
                                      self.config['length']['upload'],
                                      opener=self._opener,
                                      shutdown_event=self._shutdown_event)
                while in_flight['threads'] >= max_threads:
                    timeit.time.sleep(0.001)
                thread.start()
                q.put(thread, True)
                in_flight['threads'] += 1
                callback(i, request_count, start=True)

        finished = []

        def consumer(q, request_count):
            _is_alive = thread_is_alive
            while len(finished) < request_count:
                thread = q.get(True)
                while _is_alive(thread):
                    thread.join(timeout=0.001)
                in_flight['threads'] -= 1
                finished.append(thread.result)
                callback(thread.i, request_count, end=True)

        q = Queue(threads or self.config['threads']['upload'])
        prod_thread = threading.Thread(target=producer,
                                       args=(q, requests, request_count))
        cons_thread = threading.Thread(target=consumer,
                                       args=(q, request_count))
        start = timeit.default_timer()
        prod_thread.start()
        cons_thread.start()
        _is_alive = thread_is_alive
        while _is_alive(prod_thread):
            prod_thread.join(timeout=0.1)
        while _is_alive(cons_thread):
            cons_thread.join(timeout=0.1)

        stop = timeit.default_timer()
        self.results.bytes_sent = sum(finished)
        self.results.upload = ((self.results.bytes_sent / (stop - start)) *
                               8.0)
        return self.results.upload


def ctrl_c(shutdown_event):
    """Catch Ctrl-C key sequence and set a SHUTDOWN_EVENT for our threaded
    operations
    """
    def inner(signum, frame):
        shutdown_event.set()
        printer('\nCancelling...', error=True)
        sys.exit(0)

    return inner


def version():
    """Print the version"""

    printer('speedtest-cli %s' % __version__)
    printer('Python %s' % sys.version.replace('\n', ''))
    sys.exit(0)


def csv_header(delimiter=','):
    """Print the CSV Headers"""

    printer(SpeedtestResults.csv_header(delimiter=delimiter))
    sys.exit(0)


def parse_args():
    """Function to handle building and parsing of command line arguments"""
    description = (
        'Command line interface for testing internet bandwidth using '
        'speedtest.net.\n'
        '------------------------------------------------------------'
        '--------------\n'
        'https://github.com/sivel/speedtest-cli')

    parser = ArgParser(description=description)
    # Give optparse.OptionParser an `add_argument` method for
    # compatibility with argparse.ArgumentParser
    try:
        parser.add_argument = parser.add_option
    except AttributeError:
        pass
    parser.add_argument('--no-download',
                        dest='download',
                        default=True,
                        action='store_const',
                        const=False,
                        help='Do not perform download test')
    parser.add_argument('--no-upload',
                        dest='upload',
                        default=True,
                        action='store_const',
                        const=False,
                        help='Do not perform upload test')
    parser.add_argument('--single',
                        default=False,
                        action='store_true',
                        help='Only use a single connection instead of '
                        'multiple. This simulates a typical file '
                        'transfer.')
    parser.add_argument('--bytes',
                        dest='units',
                        action='store_const',
                        const=('byte', 8),
                        default=('bit', 1),
                        help='Display values in bytes instead of bits. Does '
                        'not affect the image generated by --share, nor '
                        'output from --json or --csv')
    parser.add_argument('--share',
                        action='store_true',
                        help='Generate and provide a URL to the speedtest.net '
                        'share results image, not displayed with --csv')
    parser.add_argument('--simple',
                        action='store_true',
                        default=False,
                        help='Suppress verbose output, only show basic '
                        'information')
    parser.add_argument('--csv',
                        action='store_true',
                        default=False,
                        help='Suppress verbose output, only show basic '
                        'information in CSV format. Speeds listed in '
                        'bit/s and not affected by --bytes')
    parser.add_argument('--csv-delimiter',
                        default=',',
                        type=PARSER_TYPE_STR,
                        help='Single character delimiter to use in CSV '
                        'output. Default ","')
    parser.add_argument('--csv-header',
                        action='store_true',
                        default=False,
                        help='Print CSV headers')
    parser.add_argument('--json',
                        action='store_true',
                        default=False,
                        help='Suppress verbose output, only show basic '
                        'information in JSON format. Speeds listed in '
                        'bit/s and not affected by --bytes')
    parser.add_argument('--list',
                        action='store_true',
                        help='Display a list of speedtest.net servers '
                        'sorted by distance')
    parser.add_argument('--server',
                        type=PARSER_TYPE_INT,
                        action='append',
                        help='Specify a server ID to test against. Can be '
                        'supplied multiple times')
    parser.add_argument('--exclude',
                        type=PARSER_TYPE_INT,
                        action='append',
                        help='Exclude a server from selection. Can be '
                        'supplied multiple times')
    parser.add_argument('--mini', help='URL of the Speedtest Mini server')
    parser.add_argument('--source', help='Source IP address to bind to')
    parser.add_argument('--timeout',
                        default=10,
                        type=PARSER_TYPE_FLOAT,
                        help='HTTP timeout in seconds. Default 10')
    parser.add_argument('--secure',
                        action='store_true',
                        help='Use HTTPS instead of HTTP when communicating '
                        'with speedtest.net operated servers')
    parser.add_argument('--no-pre-allocate',
                        dest='pre_allocate',
                        action='store_const',
                        default=True,
                        const=False,
                        help='Do not pre allocate upload data. Pre allocation '
                        'is enabled by default to improve upload '
                        'performance. To support systems with '
                        'insufficient memory, use this option to avoid a '
                        'MemoryError')
    parser.add_argument('--version',
                        action='store_true',
                        help='Show the version number and exit')
    parser.add_argument('--debug',
                        action='store_true',
                        help=ARG_SUPPRESS,
                        default=ARG_SUPPRESS)

    options = parser.parse_args()
    if isinstance(options, tuple):
        args = options[0]
    else:
        args = options
    return args


def validate_optional_args(args):
    """Check if an argument was provided that depends on a module that may
    not be part of the Python standard library.

    If such an argument is supplied, and the module does not exist, exit
    with an error stating which module is missing.
    """
    optional_args = {
        'json': ('json/simplejson python module', json),
        'secure': ('SSL support', HTTPSConnection),
    }

    for arg, info in optional_args.items():
        if getattr(args, arg, False) and info[1] is None:
            raise SystemExit('%s is not installed. --%s is '
                             'unavailable' % (info[0], arg))


def printer(string, quiet=False, debug=False, error=False, **kwargs):
    """Helper function print a string with various features"""

    if debug and not DEBUG:
        return

    if debug:
        if sys.stdout.isatty():
            out = '\033[1;30mDEBUG: %s\033[0m' % string
        else:
            out = 'DEBUG: %s' % string
    else:
        out = string

    if error:
        kwargs['file'] = sys.stderr

    if not quiet:
        print_(out, **kwargs)


def shell():
    """Run the full speedtest.net test"""

    global DEBUG
    shutdown_event = threading.Event()

    signal.signal(signal.SIGINT, ctrl_c(shutdown_event))

    args = parse_args()

    # Print the version and exit
    if args.version:
        version()

    if not args.download and not args.upload:
        raise SpeedtestCLIError('Cannot supply both --no-download and '
                                '--no-upload')

    if len(args.csv_delimiter) != 1:
        raise SpeedtestCLIError('--csv-delimiter must be a single character')

    if args.csv_header:
        csv_header(args.csv_delimiter)

    validate_optional_args(args)

    debug = getattr(args, 'debug', False)
    if debug == 'SUPPRESSHELP':
        debug = False
    if debug:
        DEBUG = True

    if args.simple or args.csv or args.json:
        quiet = True
    else:
        quiet = False

    if args.csv or args.json:
        machine_format = True
    else:
        machine_format = False

    # Don't set a callback if we are running quietly
    if quiet or debug:
        callback = do_nothing
    else:
        callback = print_dots(shutdown_event)

    printer('Retrieving speedtest.net configuration...', quiet)
    try:
        speedtest = Speedtest(source_address=args.source,
                              timeout=args.timeout,
                              secure=args.secure)
    except (ConfigRetrievalError, ) + HTTP_ERRORS:
        printer('Cannot retrieve speedtest configuration', error=True)
        raise SpeedtestCLIError(get_exception())

    if args.list:
        try:
            speedtest.get_servers()
        except (ServersRetrievalError, ) + HTTP_ERRORS:
            printer('Cannot retrieve speedtest server list', error=True)
            raise SpeedtestCLIError(get_exception())

        for _, servers in sorted(speedtest.servers.items()):
            for server in servers:
                line = ('%(id)5s) %(sponsor)s (%(name)s, %(country)s) '
                        '[%(d)0.2f km]' % server)
                try:
                    printer(line)
                except IOError:
                    e = get_exception()
                    if e.errno != errno.EPIPE:
                        raise
        sys.exit(0)

    printer('Testing from %(isp)s (%(ip)s)...' % speedtest.config['client'],
            quiet)

    if not args.mini:
        printer('Retrieving speedtest.net server list...', quiet)
        try:
            speedtest.get_servers(servers=args.server, exclude=args.exclude)
        except NoMatchedServers:
            raise SpeedtestCLIError('No matched servers: %s' %
                                    ', '.join('%s' % s for s in args.server))
        except (ServersRetrievalError, ) + HTTP_ERRORS:
            printer('Cannot retrieve speedtest server list', error=True)
            raise SpeedtestCLIError(get_exception())
        except InvalidServerIDType:
            raise SpeedtestCLIError('%s is an invalid server type, must '
                                    'be an int' %
                                    ', '.join('%s' % s for s in args.server))

        if args.server and len(args.server) == 1:
            printer('Retrieving information for the selected server...', quiet)
        else:
            printer('Selecting best server based on ping...', quiet)
        speedtest.get_best_server()
    elif args.mini:
        speedtest.get_best_server(speedtest.set_mini_server(args.mini))

    results = speedtest.results

    printer(
        'Hosted by %(sponsor)s (%(name)s) [%(d)0.2f km]: '
        '%(latency)s ms' % results.server, quiet)

    if args.download:
        printer('Testing download speed', quiet, end=('', '\n')[bool(debug)])
        speedtest.download(callback=callback, threads=(None, 1)[args.single])
        printer(
            'Download: %0.2f M%s/s' % ((results.download / 1000.0 / 1000.0) /
                                       args.units[1], args.units[0]), quiet)
    else:
        printer('Skipping download test', quiet)

    if args.upload:
        printer('Testing upload speed', quiet, end=('', '\n')[bool(debug)])
        speedtest.upload(callback=callback,
                         pre_allocate=args.pre_allocate,
                         threads=(None, 1)[args.single])
        printer(
            'Upload: %0.2f M%s/s' % ((results.upload / 1000.0 / 1000.0) /
                                     args.units[1], args.units[0]), quiet)
    else:
        printer('Skipping upload test', quiet)

    printer('Results:\n%r' % results.dict(), debug=True)

    if not args.simple and args.share:
        results.share()

    if args.simple:
        printer('Ping: %s ms\nDownload: %0.2f M%s/s\nUpload: %0.2f M%s/s' %
                (results.ping, (results.download / 1000.0 / 1000.0) /
                 args.units[1], args.units[0],
                 (results.upload / 1000.0 / 1000.0) / args.units[1],
                 args.units[0]))
    elif args.csv:
        printer(results.csv(delimiter=args.csv_delimiter))
    elif args.json:
        printer(results.json())

    if args.share and not machine_format:
        printer('Share results: %s' % results.share())


def main():
    try:
        shell()
    except KeyboardInterrupt:
        printer('\nCancelling...', error=True)
    except (SpeedtestException, SystemExit):
        e = get_exception()
        # Ignore a successful exit, or argparse exit
        if getattr(e, 'code', 1) not in (0, 2):
            msg = '%s' % e
            if not msg:
                msg = '%r' % e
            raise SystemExit('ERROR: %s' % msg)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
import platform
import subprocess
import sys

import requests

url_log = ~/.url_log.log

def getURL():
    print(platform.platform())

    """ getURL() {
            # copy URL from active tab of Google Chrome to macOS clipboard
            if [$OS= 'darwin']
            then
            url =$(osascript - e 'tell application "Google Chrome" to return URL of active tab of front window')
            # printf "%s" "$url" | pbcopy # skip clipboard on -v

            # -v is verbose : print url to stdout in addition to clipboard
            if ["$1" = '-v']
            then
            printf "%s\n" "$url"
            else
            printf "%s" "$url" | pbcopy
            fi
            fi
        }
        """


def main():
    '''
    CLI script main entry point.
    '''
    getURL()


if __name__ == "__main__":  # if script is loaded directly from CLI
    main()


"""
ping_avg() {
    # extract the average ping time (ms) from 'ping' output
    # $1 = number of pings to average
    # $2 = url to test
    count =$1
    url =$(echo "$2" | cut - d '/' - f 3)  # remove schema and folders
    ping - c "$count" "$url" | tail - n 1 | cut - d '/' - f 6
}
"""
"""
log_urls() {
    # ref: https://www.cyberciti.biz/faq/linux-unix-sleep-bash-scripting/
    url_log = ~/.url_log.log
    # shellcheck disable=SC2078
    while [:]
    do
    url =$(getURL - v)
    # >>$url_log
    printf "%s\n" "$(date), $url, $(ping_avg 2 $url), $(ping_avg 2 'www.example.com')"
    printf "%s\n" "$(date), $url, $(ping_avg 2 $url), $(ping_avg 2 'www.example.com')" >>$url_log
    sleep 10
    done
}
"""



# Python Tools for Visual Studio
# Copyright(c) Microsoft Corporation
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the License); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0
#
# THIS CODE IS PROVIDED ON AN  *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
# IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
# MERCHANTABLITY OR NON-INFRINGEMENT.
#
# See the Apache Version 2.0 License for specific language governing
# permissions and limitations under the License.

__author__ = "Microsoft Corporation <ptvshelp@microsoft.com>"
__version__ = "3.2"

# 'Standard Library'
import os
import site
import sys
import zipfile

# HACK: macOS sets __cached__ to None for __main__
# but site.main() cannot handle it. So we force it to a str.
__cached__ = ""

if "site" in sys.modules:
    raise RuntimeError("script must be run with -S")

BEFORE_SITE = list(sys.path)


try:
    site.main()
except:
    import traceback

    traceback.print_exc(file=sys.stderr)
AFTER_SITE = list(sys.path)



def clean(path):
    if path:
        return os.path.normcase(os.path.abspath(path).rstrip(os.sep))
    return None


BEFORE_SITE = set(clean(p) for p in BEFORE_SITE)
AFTER_SITE = set(clean(p) for p in AFTER_SITE)
SCRIPT_DIR = clean(os.path.dirname(os.path.realpath(__file__)))

try:
    SITE_PKGS = set(clean(p) for p in site.getsitepackages())
except AttributeError:
    SITE_PKGS = set()

for prefix in [
    sys.prefix,
    sys.exec_prefix,
    getattr(sys, "real_prefix", ""),
    getattr(sys, "base_prefix", ""),
]:
    if not prefix:
        continue

    BEFORE_SITE.add(clean(prefix))

    for subdir in ["DLLs", "Lib", "Scripts"]:
        d = clean(os.path.join(prefix, subdir))
        BEFORE_SITE.add(d)

BEFORE_SITE.discard(None)
AFTER_SITE.discard(None)


for p in sys.path:
    p = clean(p)

    if p == SCRIPT_DIR or p.startswith(SCRIPT_DIR + os.sep):
        continue

    if not os.path.isdir(p) and not (
        os.path.isfile(p) and zipfile.is_zipfile(p)
    ):
        continue

    if p in BEFORE_SITE:
        print("%s|stdlib|" % p)
    elif p in AFTER_SITE:
        if p in SITE_PKGS:
            print("%s|site|" % p)
        else:
            print("%s|pth|" % p)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# 'future imports'
from __future__ import absolute_import

# 'Standard Library'
import sys

from collections import deque
from contextlib import closing
from dataclasses import dataclass
from os import environ as ENV, linesep as NL
from pprint import pprint
from sys import argv, stderr, stdout

# 'package imports'
from autosys.web.web_basics import *

from bs4 import BeautifulSoup
# 'third party'
from requests import get
from requests.exceptions import RequestException

from typing import Any, Deque, Dict, List, Tuple

""" scrape.py - web scraping utilities for python
    (reference: https://realpython.com/python-web-scraping-practical-introduction/)

        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal
    """

try:
    import lxml as parser

    DEFAULT_PARSER = "lxml"
except:
    DEFAULT_PARSER = "html.parser"

# !---------------------------------------------- Custom Types


@dataclass(frozen=True)
class ScrapeRules(List[Any]):
    """ A list of functions, regex matches, or other checks to perform
        that keep the web scraper from doing anything unwanted. """

    triggers: List[Any] = []


@dataclass()
class WebDocument:
    """ Manages a List of additional url information for webpages.
        - Places the url in a data structure with additional optional fields
        - Allows automated recording of data during the web scraping process.
        - Any method that begins with '__wd_auto_' runs on url access
        - Add additional temporary methods with decorators
        - Add additional permanent methods with subclassing
        - Each method can be toggled on or off individually
        - Performance information for each addon is recorded
        - Gathers data that would otherwise be lost.
        - the input and output can remain a simple url.
        - each url may have multiple access and generate an audit trail

        Examples of optional fields. Add others by subclassing.
        - date and time of web access
        - author and contributor names and info
        - info required for citations
        - ping and download times
        - current user / machine / os / user-agent / environment
        - DNS server and proxy information
        - redirections encountered
        - additional unplanned data packets (is data encapsulated?)
        - complete status information (instead of just 200==success)
        - location of further resources (nonlocal links / images)
        - generate a list of further resources as a script
        """

    url: str


@dataclass
class ScrapeDocument(WebDocument):
    scrape_rules: List[Any] = []


@dataclass()
class ScrapeDocumentSet:
    url_list: List[str] = []
    config: ConfigDefaults = ConfigDefaults()
    DEFAULT_URL_HISTORY: int = 10000
    _url_set: Deque = deque(url_list, maxlen=DEFAULT_URL_HISTORY)
    optional_fields: List[Any] = WD_AUTO_STANDARD_FIELDS
    scrape_rules_list: List[Any] = DEFAULT_SCRAPE_RULES

    def __append__(self, url):
        self._url_set.__append__(
            ScrapeDocument(url=url, scrape_rules=self.scrape_rules_list))


# !---------------------------------------------- Script Tests
def __tests__(args) -> int:
    """ Run Debug Tests for script if _debug_ = True. """

    print(f"{DEFAULT_PARSER=}")
    url = "https://realpython.com/blog/"
    raw_html = simple_get(url)
    html = soup(url)
    pprint(html)
    return 0


def __main__(args) -> int:
    """ CLI script main entry point. """

    #! script testing
    if _debug_:
        __tests__(args)
    return 0


if __name__ == "__main__":  # if script is loaded directly from CLI
    __main__(sys.argv[1:])
""" Notes
    from tutorial ... moved out of __test__()

    raw_html = simple_get("https://realpython.com/blog/")
    print(len(raw_html))

    no_html = simple_get("https://realpython.com/blog/nope-not-gonna-find-it")

    print(no_html is None)

    # raw_html = open("contrived.html").read()
    """



# a quirky import to allow temporary (quick and dirty) debugging
_debug_: bool = True
START: str = "="
MID: str = "-"
END: str = "#"
REPEAT: int = 30


def _debug_it(func):
    global _debug_

    def _debug_it_(*a, **kw):
        global _debug_
        print(START * REPEAT)
        print(f"{a=},  {kw=}")
        try:
            _debug_ = bool(kw.pop("debug"))
            print(f"set _debug_ to: {_debug_}")
        except:
            _debug_ = False
        print(f"{_debug_=}")
        print(a, kw)
        print(MID * REPEAT)
        result = func(*a, **kw)
        print(MID * REPEAT)
        print(f"{_debug_=}")
        print(END * REPEAT)
        return result

    return _debug_it_


@_debug_it
def stuff(number=0, *args, **kw):
    print(f"function number: {number}")
    print(f"  ==> {_debug_=}")
    return number * 5


__all__ = ["_debug_", "_debug_it"]


def _test():
    print(stuff(1, "number 1", debug=False))
    print(stuff(0, "number 0", debug=True))
    print(stuff(1, "number 1", debug=5))
    print(stuff(2, "number2", debug=True))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Downloads all tweets from a given user.

Uses twitter.Api.GetUserTimeline to retreive the last 3,200 tweets from a user.
Twitter doesn't allow retreiving more tweets than this through the API, so we get
as many as possible.

t.py should contain the imported variables.
"""

# 'future imports'
from __future__ import print_function

# 'Standard Library'
import json
import sys

# 'package imports'
import twitter

from t import (
    ACCESS_TOKEN_KEY, ACCESS_TOKEN_SECRET,
    CONSUMER_KEY, CONSUMER_SECRET)


def get_tweets(api=None, screen_name=None):
    timeline = api.GetUserTimeline(screen_name=screen_name, count=200)
    earliest_tweet = min(timeline, key=lambda x: x.id).id
    print("getting tweets before:", earliest_tweet)

    while True:
        tweets = api.GetUserTimeline(
            screen_name=screen_name, max_id=earliest_tweet, count=200
        )
        new_earliest = min(tweets, key=lambda x: x.id).id

        if not tweets or new_earliest == earliest_tweet:
            break
        else:
            earliest_tweet = new_earliest
            print("getting tweets before:", earliest_tweet)
            timeline += tweets

    return timeline


if __name__ == "__main__":
    api = twitter.Api(
        CONSUMER_KEY, CONSUMER_SECRET, ACCESS_TOKEN_KEY, ACCESS_TOKEN_SECRET
    )
    screen_name = sys.argv[1]
    print(screen_name)
    timeline = get_tweets(api=api, screen_name=screen_name)

    with open("examples/timeline.json", "w+") as f:
        for tweet in timeline:
            f.write(json.dumps(tweet._json))
            f.write("\n")



# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
__license__ = "MIT"

# * text messaging
DEFAULT_SCRIPT_PATH: str = os.getcwd() + "/sentMessage.scpt"
DEFAULT_CELL_NUMBER: str = "13616488261"


class Text_Messaging:
    def text(
        self,
        message: str,
        cell: str = DEFAULT_CELL_NUMBER,
        verbose: bool = False,
    ) -> int:
        """ Return status of {message} sent to {cell}(verbose - give CLI feedback if True """
        script_text = "osascript {} {} '{}'".format(
            DEFAULT_SCRIPT_PATH, cell, message
        )
        db_print("script_text: ", script_text)
        try:
            result = os.system(script_text)
        except:
            result = -1
        finally:
            db_print("result: ", result)
            return result

    def test_text(
        self, message: str = "A python program just sent you a message ..."
    ):
        """ Send a test message using <text> """
        test_cell_number = DEFAULT_CELL_NUMBER
        db_print(f"{message=}")
        result = text(message=message, verbose=SET_DEBUG)
        print(f"{result=}")



def __getattr__(name):
    import importlib

    if name in __all__:
        return importlib.import_module("." + name, __name__)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def get_class_name(_name, verbose=True):
    """ Alternate class name function with optional formatting """
    _cls_name = eval(f"{_name}.__class__.__name__")
    if not verbose:
        return _cls_name
    return f"{_name} - <class '{_cls_name}'>"


# generic test class for `add_method`
A = type("A", (object,), dict(a=1))


def add_method(cls):
    """ Decorator to add method to class 'cls' """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(*args, **kwargs)

        setattr(cls, func.__name__, wrapper)
        # Note we are not binding func, but wrapper which accepts self but does exactly the same as func
        return func  # returning func means func can still be used normally

    return decorator


# Samples


# Decorator can be written to take normal functions and make them methods. These functions still act as normal if passed a standard variable. If they are passed a method parameter, they are run as class methods with 'self' injected in as needed
@add_method(A)
def foo():
    print("hello world!")


@add_method(A)
def bar(s):
    print(f"Message: {s}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# 'package imports'
from dev.debug import *

# !---------------------------------------------- dev tests


def _tests_():
    x = ""
    print(x)


# !---------------------------------------------- cli features


def _main_():
    """
    CLI script main entry point.
    """
    _debug_: bool = True
    if _debug_:
        _tests_()


if __name__ == "__main__":  # if script is loaded directly from CLI
    _main_()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" QuickColors Setup
    ---
    Part of the [AutoSys][1] package

    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """

import setup_utils

setup_utils
